# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""MCP client."""

import abc
from typing import Annotated, Type

from langfun.core.mcp import session as mcp_session
from langfun.core.mcp import tool as mcp_tool
from mcp.server import fastmcp as fastmcp_lib
import pyglove as pg


class McpClient(pg.Object):
  """Base class for MCP client.

  Usage:

  ```python

  def tool_use():
    client = lf.mcp.McpClient.from_command('<MCP_CMD>', ['<ARG1>', 'ARG2'])
    tools = client.list_tools()
    tool_cls = tools['<TOOL_NAME>']

    # Print the python definition of the tool.
    print(tool_cls.python_definition())

    with client.session() as session:
      return tool_cls(x=1, y=2)(session)

  async def tool_use_async_version():
    client = lf.mcp.McpClient.from_url('http://localhost:8000/mcp')
    tools = client.list_tools()
    tool_cls = tools['<TOOL_NAME>']

    # Print the python definition of the tool.
    print(tool_cls.python_definition())

    async with client.session() as session:
      return await tool_cls(x=1, y=2).acall(session)
  ```
  """

  def _on_bound(self):
    super()._on_bound()
    self._tools = None

  def list_tools(
      self, refresh: bool = False
  ) -> dict[str, Type[mcp_tool.McpTool]]:
    """Lists all MCP tools."""
    if self._tools is None or refresh:
      with self.session() as session:
        self._tools = session.list_tools()
    return self._tools

  @abc.abstractmethod
  def session(self) -> mcp_session.McpSession:
    """Creates a MCP session."""

  @classmethod
  def from_command(cls, command: str, args: list[str]) -> 'McpClient':
    """Creates a MCP client from a tool."""
    return _StdioMcpClient(command=command, args=args)

  @classmethod
  def from_url(
      cls,
      url: str,
      headers: dict[str, str] | None = None
  ) -> 'McpClient':
    """Creates a MCP client from a URL."""
    return _HttpMcpClient(url=url, headers=headers or {})

  @classmethod
  def from_fastmcp(cls, fastmcp: fastmcp_lib.FastMCP) -> 'McpClient':
    """Creates a MCP client from a MCP server."""
    return _InMemoryFastMcpClient(fastmcp=fastmcp)


class _StdioMcpClient(McpClient):
  """Stdio-based MCP client."""

  command: Annotated[str, 'Command to execute.']
  args: Annotated[list[str], 'Arguments to pass to the command.']

  def session(self) -> mcp_session.McpSession:
    """Creates a MCP session."""
    return mcp_session.McpSession.from_command(self.command, self.args)


class _HttpMcpClient(McpClient):
  """Server-Sent Events (SSE)/Streamable HTTP-based MCP client."""

  url: Annotated[str, 'URL to connect to.']
  headers: Annotated[dict[str, str], 'Headers to send with the request.'] = {}

  def session(self) -> mcp_session.McpSession:
    """Creates a MCP session."""
    return mcp_session.McpSession.from_url(self.url, self.headers)


class _InMemoryFastMcpClient(McpClient):
  """In-memory MCP client."""

  fastmcp: Annotated[fastmcp_lib.FastMCP, 'MCP server to connect to.']

  def session(self) -> mcp_session.McpSession:
    """Creates a MCP session."""
    return mcp_session.McpSession.from_fastmcp(self.fastmcp)



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import unittest
from langfun.env import interface


class IdTest(unittest.TestCase):

  def test_environment_id(self):
    env_id = interface.Environment.Id('env@1/a b:c#def')
    self.assertEqual(str(env_id), 'env@1/a b:c#def')
    self.assertEqual(
        env_id.working_dir(root_dir='/tmp'),
        '/tmp/env_1/ab_c_def'
    )
    self.assertIsNone(env_id.working_dir(root_dir=None))

  def test_sandbox_id(self):
    sandbox_id = interface.Sandbox.Id(
        environment_id=interface.Environment.Id('env'),
        sandbox_id='sandbox'
    )
    self.assertEqual(str(sandbox_id), 'env/sandbox')
    self.assertEqual(
        sandbox_id.working_dir(root_dir='/tmp'),
        '/tmp/env/sandbox'
    )
    self.assertIsNone(sandbox_id.working_dir(root_dir=None))


if __name__ == '__main__':
  unittest.main()



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import time
from typing import Any
import unittest

from langfun.env import base_sandbox
from langfun.env import interface
from langfun.env import test_utils
from langfun.env.event_handlers import base as event_handler_base


TestingEnvironment = test_utils.TestingEnvironment
TestingSandbox = test_utils.TestingSandbox
TestingFeature = test_utils.TestingFeature
TestingEventHandler = test_utils.TestingEventHandler


class EnvironmentTests(unittest.TestCase):

  def test_basics(self):
    env = TestingEnvironment(
        root_dir='/tmp',
        pool_size=0,
        features={'test_feature': TestingFeature()},
        outage_grace_period=1,
        outage_retry_interval=0,
    )
    self.assertIsNone(interface.Environment.current())
    self.assertEqual(env.status, interface.Environment.Status.CREATED)
    self.assertFalse(env.is_online)
    self.assertEqual(env.min_pool_size, 0)
    self.assertEqual(env.max_pool_size, 0)
    self.assertEqual(env.sandbox_pool, [])
    self.assertEqual(env.id, interface.Environment.Id('testing-env'))
    self.assertEqual(env.outage_grace_period, 1)
    self.assertEqual(env.features['test_feature'].name, 'test_feature')

    self.assertIsNone(env.start_time)

    with env:
      self.assertEqual(env.status, interface.Environment.Status.ONLINE)
      self.assertIs(interface.Environment.current(), env)
      self.assertTrue(env.is_online)
      self.assertIsNotNone(env.start_time)
      self.assertEqual(env.offline_duration, 0.0)
      self.assertEqual(env.sandbox_pool, [])
      self.assertEqual(env.working_dir, '/tmp/testing-env')

      with env.sandbox('session1') as sb:
        self.assertEqual(
            sb.id, interface.Sandbox.Id(environment_id=env.id, sandbox_id='0')
        )
        self.assertEqual(sb.session_id, 'session1')
        self.assertEqual(sb.working_dir, '/tmp/testing-env/0')
        self.assertTrue(sb.is_online)
        self.assertIs(sb.test_feature, sb.features['test_feature'])
        self.assertEqual(
            sb.test_feature.working_dir,
            '/tmp/testing-env/0/test_feature'
        )
        with self.assertRaises(AttributeError):
          _ = sb.test_feature2
      self.assertFalse(sb.is_online)

      self.assertIsInstance(env.test_feature, TestingFeature)
      with self.assertRaises(AttributeError):
        _ = env.test_feature2

  def test_del(self):
    env = TestingEnvironment(
        features={'test_feature': TestingFeature()},
        pool_size=0,
        outage_grace_period=1,
        outage_retry_interval=0,
        sandbox_keepalive_interval=0,
    )
    env.start()
    sb = env.acquire()
    del sb
    del env

  def test_acquire_env_offline(self):
    env = TestingEnvironment(
        features={'test_feature': TestingFeature()},
        pool_size=0,
        outage_grace_period=1,
        outage_retry_interval=0,
        sandbox_keepalive_interval=0,
    )
    with self.assertRaises(interface.EnvironmentOutageError):
      env.acquire()

  def test_acquire_no_pooling(self):
    env = TestingEnvironment(
        features={'test_feature': TestingFeature()},
        pool_size=0,
        outage_grace_period=1,
        outage_retry_interval=0,
        sandbox_keepalive_interval=0,
    )
    with env:
      sb = env.acquire()
      self.assertEqual(sb.status, interface.Sandbox.Status.ACQUIRED)
      self.assertIsNone(env.working_dir)
      self.assertIsNone(sb.working_dir)
      self.assertIsNone(sb.test_feature.working_dir)

  def test_acquire_no_pooling_with_error(self):
    env = TestingEnvironment(
        features={
            'test_feature': TestingFeature(
                simulate_setup_error=interface.SandboxStateError
            )
        },
        pool_size=0,
        outage_grace_period=1,
        outage_retry_interval=0,
        sandbox_keepalive_interval=0,
    )
    with env:
      with self.assertRaises(interface.EnvironmentOutageError):
        env.acquire()

  def test_acquire_with_pooling(self):
    env = TestingEnvironment(
        features={'test_feature': TestingFeature()},
        pool_size=1,
        outage_grace_period=1,
        outage_retry_interval=0,
        sandbox_keepalive_interval=0,
    )
    with env:
      sb = env.acquire()
      self.assertEqual(sb.status, interface.Sandbox.Status.ACQUIRED)

  def test_acquire_with_pooling_overload(self):
    env = TestingEnvironment(
        features={'test_feature': TestingFeature()},
        pool_size=1,
        outage_grace_period=1,
        outage_retry_interval=0,
        sandbox_keepalive_interval=0,
    )
    with env:
      sb = env.acquire()
      self.assertEqual(sb.status, interface.Sandbox.Status.ACQUIRED)
      with self.assertRaises(interface.EnvironmentOverloadError):
        env.acquire()

  def test_acquire_with_growing_pool(self):
    env = TestingEnvironment(
        features={'test_feature': TestingFeature()},
        pool_size=(1, 3),
        outage_grace_period=1,
        outage_retry_interval=0,
        sandbox_keepalive_interval=0,
    )
    with env:
      self.assertEqual(len(env.sandbox_pool), 1)
      self.assertEqual(
          env.stats(),
          {
              'sandbox': {
                  'created': 0,
                  'setting_up': 0,
                  'ready': 1,
                  'acquired': 0,
                  'in_session': 0,
                  'exiting_session': 0,
                  'shutting_down': 0,
                  'offline': 0,
              }
          }
      )
      sb = env.acquire()
      self.assertEqual(sb.status, interface.Sandbox.Status.ACQUIRED)
      self.assertEqual(
          env.stats(),
          {
              'sandbox': {
                  'created': 0,
                  'setting_up': 0,
                  'ready': 0,
                  'acquired': 1,
                  'in_session': 0,
                  'exiting_session': 0,
                  'shutting_down': 0,
                  'offline': 0,
              }
          }
      )
      self.assertEqual(len(env.sandbox_pool), 1)
      sb2 = env.acquire()
      self.assertEqual(sb2.status, interface.Sandbox.Status.ACQUIRED)
      self.assertEqual(len(env.sandbox_pool), 2)
      self.assertEqual(
          env.stats(),
          {
              'sandbox': {
                  'created': 0,
                  'setting_up': 0,
                  'ready': 0,
                  'acquired': 2,
                  'in_session': 0,
                  'exiting_session': 0,
                  'shutting_down': 0,
                  'offline': 0,
              }
          }
      )
    self.assertEqual(
        env.stats(),
        {
            'sandbox': {
                'created': 0,
                'setting_up': 0,
                'ready': 0,
                'acquired': 0,
                'in_session': 0,
                'exiting_session': 0,
                'shutting_down': 0,
                'offline': 0,
            }
        }
    )

  def test_acquire_with_growing_pool_failure(self):
    env = TestingEnvironment(
        features={'test_feature': TestingFeature()},
        pool_size=(1, 3),
        outage_grace_period=1,
        outage_retry_interval=0,
        sandbox_keepalive_interval=0,
    )
    with env:
      self.assertEqual(len(env.sandbox_pool), 1)
      sb = env.acquire()
      self.assertEqual(sb.status, interface.Sandbox.Status.ACQUIRED)

      # Make future sandbox setup to fail.
      env.features.test_feature.rebind(
          simulate_setup_error=interface.SandboxStateError,
          skip_notification=True
      )
      with self.assertRaises(interface.EnvironmentOutageError):
        env.acquire()

  def test_housekeep_error(self):
    env = TestingEnvironment(
        features={'test_feature': TestingFeature()},
        pool_size=1,
        proactive_session_setup=True,
        outage_grace_period=1,
        outage_retry_interval=0,
        sandbox_keepalive_interval=0,
    )
    with env:
      self.assertEqual(len(env.sandbox_pool), 1)
      self.assertEqual(
          env.sandbox_pool[0].status, interface.Sandbox.Status.READY
      )
      # Make future sandbox setup to fail.
      env.features.test_feature.rebind(
          simulate_setup_error=interface.SandboxStateError,
          skip_notification=True
      )
      with env.sandbox() as sb:
        with self.assertRaises(interface.SandboxStateError):
          sb.shell('bad command', raise_error=interface.SandboxStateError)
      self.assertEqual(sb.status, interface.Sandbox.Status.OFFLINE)
      sb_offline_time = time.time()
      while time.time() - sb_offline_time < 10:
        if not env.is_online:
          break
        time.sleep(0.5)
      self.assertFalse(env.is_online)


class SandboxStatusTests(unittest.TestCase):

  def setUp(self):
    super().setUp()
    self.event_handler = TestingEventHandler(
        log_sandbox_status=True,
        log_feature_setup=True,
        log_session_setup=True,
    )
    self.maxDiff = None

  def _create_env(
      self,
      features,
      *,
      pool_size=0,
      **kwargs
  ) -> TestingEnvironment:
    return TestingEnvironment(
        pool_size=pool_size,
        features=features,
        outage_grace_period=0,
        event_handlers=[self.event_handler],
        outage_retry_interval=0,
        **kwargs
    )

  def test_passive_session_setup(self):
    env = self._create_env(
        features={
            'feature1': TestingFeature(),
            'feature2': TestingFeature(),
        },
    )
    self.assertFalse(env.enable_pooling)
    with env:
      with env.sandbox('session1') as sb:
        sb.shell('echo "hello"')
      self.assertEqual(
          self.event_handler.logs,
          [
              '[testing-env] environment started',
              '[testing-env/0] shell: "feature1" setup',
              '[testing-env/0/feature1] feature setup',
              '[testing-env/0] shell: "feature2" setup',
              '[testing-env/0/feature2] feature setup',
              '[testing-env/0] created -> ready',
              '[testing-env/0] sandbox started',
              '[testing-env/0] ready -> acquired',
              '[testing-env/0] acquired -> setting_up',
              '[testing-env/0/session1] shell: "feature1" setup session',
              '[testing-env/0/feature1] feature setup session',
              '[testing-env/0/session1] shell: "feature2" setup session',
              '[testing-env/0/feature2] feature setup session',
              '[testing-env/0] setting_up -> in_session',
              "[testing-env/0] session 'session1' started",
              '[testing-env/0/session1] shell: echo "hello"',
              '[testing-env/0] in_session -> exiting_session',
              '[testing-env/0/session1] shell: "feature1" teardown session',
              '[testing-env/0/feature1] feature teardown session',
              '[testing-env/0/session1] shell: "feature2" teardown session',
              '[testing-env/0/feature2] feature teardown session',
              "[testing-env/0] session 'session1' ended",
              '[testing-env/0] exiting_session -> acquired',
              '[testing-env/0] acquired -> shutting_down',
              '[testing-env/0] shell: "feature1" teardown',
              '[testing-env/0/feature1] feature teardown',
              '[testing-env/0] shell: "feature2" teardown',
              '[testing-env/0/feature2] feature teardown',
              '[testing-env/0] shutting_down -> offline',
              '[testing-env/0] sandbox shutdown'
          ]
      )

  def test_proactive_session_setup(self):
    env = self._create_env(
        features={
            'feature1': TestingFeature(setup_session_delay=0.1),
            'feature2': TestingFeature(),
        },
        pool_size=1,
        proactive_session_setup=True,
    )
    self.assertTrue(env.enable_pooling)
    with env:
      with env.sandbox('session1') as sb:
        sb.shell('echo "hello"')
      sb.wait_until_not(
          (
              interface.Sandbox.Status.IN_SESSION,
              interface.Sandbox.Status.SETTING_UP
          )
      )
      self.assertEqual(sb.status, interface.Sandbox.Status.READY)
      self.assertEqual(
          self.event_handler.logs,
          [
              '[testing-env/0:0] shell: "feature1" setup',
              '[testing-env/0:0/feature1] feature setup',
              '[testing-env/0:0] shell: "feature2" setup',
              '[testing-env/0:0/feature2] feature setup',
              '[testing-env/0:0] shell: "feature1" setup session',
              '[testing-env/0:0/feature1] feature setup session',
              '[testing-env/0:0] shell: "feature2" setup session',
              '[testing-env/0:0/feature2] feature setup session',
              '[testing-env/0:0] created -> ready',
              '[testing-env/0:0] sandbox started',
              '[testing-env] environment started',
              '[testing-env/0:0] ready -> acquired',
              '[testing-env/0:0] acquired -> setting_up',
              '[testing-env/0:0] setting_up -> in_session',
              "[testing-env/0:0] session 'session1' started",
              '[testing-env/0:0/session1] shell: echo "hello"',
              '[testing-env/0:0] in_session -> exiting_session',
              '[testing-env/0:0/session1] shell: "feature1" teardown session',
              '[testing-env/0:0/feature1] feature teardown session',
              '[testing-env/0:0/session1] shell: "feature2" teardown session',
              '[testing-env/0:0/feature2] feature teardown session',
              "[testing-env/0:0] session 'session1' ended",
              '[testing-env/0:0] exiting_session -> setting_up',
              '[testing-env/0:0] shell: "feature1" setup session',
              '[testing-env/0:0/feature1] feature setup session',
              '[testing-env/0:0] shell: "feature2" setup session',
              '[testing-env/0:0/feature2] feature setup session',
              '[testing-env/0:0] setting_up -> ready'
          ]
      )

  def test_practive_session_setup_with_setup_session_error(self):
    env = self._create_env(
        features={'test_feature': TestingFeature(setup_session_delay=0.5)},
        pool_size=1,
    )
    event_handler = TestingEventHandler(
        log_sandbox_status=True,
        log_feature_setup=True,
        log_session_setup=True,
    )
    with env:
      with env.sandbox('session1') as sb:
        sb.add_event_handler(event_handler)
        sb.test_feature.rebind(
            simulate_setup_session_error=interface.SandboxStateError,
            skip_notification=True
        )
      sb.wait_until_not(
          (
              interface.Sandbox.Status.SETTING_UP,
              interface.Sandbox.Status.SHUTTING_DOWN
          )
      )
      self.assertEqual(len(sb.state_errors), 1)
      self.assertEqual(sb.status, interface.Sandbox.Status.OFFLINE)
      self.assertEqual(
          event_handler.logs,
          [
              # pylint: disable=line-too-long
              '[testing-env/0:0] in_session -> exiting_session',
              '[testing-env/0:0/session1] shell: "test_feature" teardown session',
              '[testing-env/0:0/test_feature] feature teardown session',
              "[testing-env/0:0] session 'session1' ended",
              '[testing-env/0:0] exiting_session -> setting_up',
              '[testing-env/0:0/test_feature] feature setup session with SandboxStateError',  # pylint: disable=line-too-long
              '[testing-env/0:0] setting_up -> shutting_down',
              '[testing-env/0:0] shell: "test_feature" teardown',
              '[testing-env/0:0/test_feature] feature teardown',
              '[testing-env/0:0] shutting_down -> offline',
              '[testing-env/0:0] sandbox shutdown'
              # pylint: enable=line-too-long
          ]
      )

  def test_sandbox_start_non_state_error(self):
    env = self._create_env(
        features={
            'feature1': TestingFeature(),
            'feature2': TestingFeature(),
        },
        simulate_start_error=ValueError,
    )
    with env:
      with self.assertRaises(ValueError):
        with env.sandbox('session1'):
          pass
      self.assertTrue(env.is_online)
      self.assertEqual(
          self.event_handler.logs,
          [
              '[testing-env] environment started',
              '[testing-env/0] sandbox started with ValueError',
              '[testing-env/0] created -> shutting_down',
              '[testing-env/0] shutting_down -> offline',
              '[testing-env/0] sandbox shutdown'
          ]
      )

  def test_sandbox_start_state_error(self):
    env = self._create_env(
        features={
            'feature1': TestingFeature(),
            'feature2': TestingFeature(),
        },
        pool_size=1,
        simulate_start_error=interface.SandboxStateError,
    )
    with self.assertRaises(interface.EnvironmentOutageError):
      with env:
        pass
    self.assertEqual(
        self.event_handler.logs,
        [
            '[testing-env/0:0] sandbox started with SandboxStateError',
            '[testing-env/0:0] created -> shutting_down',
            '[testing-env/0:0] shutting_down -> offline',
            '[testing-env/0:0] sandbox shutdown',
            '[testing-env] environment started with EnvironmentOutageError',
            '[testing-env] environment shutdown'
        ]
    )

  def test_sandbox_shutdown_non_state_error(self):
    env = self._create_env(
        features={
            'feature1': TestingFeature(),
            'feature2': TestingFeature(),
        },
        simulate_shutdown_error=ValueError,
    )
    with env:
      with self.assertRaises(ValueError):
        with env.sandbox('session1') as sb:
          sb.shell('echo "hello"')
      self.assertEqual(len(sb.state_errors), 0)

    self.assertEqual(
        self.event_handler.logs,
        [
            '[testing-env] environment started',
            '[testing-env/0] shell: "feature1" setup',
            '[testing-env/0/feature1] feature setup',
            '[testing-env/0] shell: "feature2" setup',
            '[testing-env/0/feature2] feature setup',
            '[testing-env/0] created -> ready',
            '[testing-env/0] sandbox started',
            '[testing-env/0] ready -> acquired',
            '[testing-env/0] acquired -> setting_up',
            '[testing-env/0/session1] shell: "feature1" setup session',
            '[testing-env/0/feature1] feature setup session',
            '[testing-env/0/session1] shell: "feature2" setup session',
            '[testing-env/0/feature2] feature setup session',
            '[testing-env/0] setting_up -> in_session',
            "[testing-env/0] session 'session1' started",
            '[testing-env/0/session1] shell: echo "hello"',
            '[testing-env/0] in_session -> exiting_session',
            '[testing-env/0/session1] shell: "feature1" teardown session',
            '[testing-env/0/feature1] feature teardown session',
            '[testing-env/0/session1] shell: "feature2" teardown session',
            '[testing-env/0/feature2] feature teardown session',
            "[testing-env/0] session 'session1' ended",
            '[testing-env/0] exiting_session -> acquired',
            '[testing-env/0] acquired -> shutting_down',
            '[testing-env/0] shell: "feature1" teardown',
            '[testing-env/0/feature1] feature teardown',
            '[testing-env/0] shell: "feature2" teardown',
            '[testing-env/0/feature2] feature teardown',
            '[testing-env/0] shutting_down -> offline',
            '[testing-env/0] sandbox shutdown with ValueError',
            '[testing-env] environment shutdown'
        ]
    )

  def test_env_shutdown_non_state_error(self):
    env = self._create_env(
        pool_size=1,
        features={
            'feature1': TestingFeature(),
            'feature2': TestingFeature(),
        },
        simulate_shutdown_error=ValueError,
    )
    with self.assertRaises(ValueError):
      with env:
        pass

    self.assertEqual(
        self.event_handler.logs,
        [
            '[testing-env/0:0] shell: "feature1" setup',
            '[testing-env/0:0/feature1] feature setup',
            '[testing-env/0:0] shell: "feature2" setup',
            '[testing-env/0:0/feature2] feature setup',
            '[testing-env/0:0] shell: "feature1" setup session',
            '[testing-env/0:0/feature1] feature setup session',
            '[testing-env/0:0] shell: "feature2" setup session',
            '[testing-env/0:0/feature2] feature setup session',
            '[testing-env/0:0] created -> ready',
            '[testing-env/0:0] sandbox started',
            '[testing-env] environment started',
            '[testing-env/0:0] ready -> shutting_down',
            '[testing-env/0:0] shell: "feature1" teardown',
            '[testing-env/0:0/feature1] feature teardown',
            '[testing-env/0:0] shell: "feature2" teardown',
            '[testing-env/0:0/feature2] feature teardown',
            '[testing-env/0:0] shutting_down -> offline',
            '[testing-env/0:0] sandbox shutdown with ValueError',
            '[testing-env] environment shutdown with ValueError'
        ]
    )

  def test_sandbox_shutdown_state_error(self):
    env = self._create_env(
        features={
            'feature1': TestingFeature(),
            'feature2': TestingFeature(),
        },
        simulate_shutdown_error=interface.SandboxStateError,
    )
    with env:
      with env.sandbox('session1') as sb:
        sb.shell('echo "hello"')
      self.assertEqual(len(sb.state_errors), 1)
    self.assertEqual(
        self.event_handler.logs,
        [
            '[testing-env] environment started',
            '[testing-env/0] shell: "feature1" setup',
            '[testing-env/0/feature1] feature setup',
            '[testing-env/0] shell: "feature2" setup',
            '[testing-env/0/feature2] feature setup',
            '[testing-env/0] created -> ready',
            '[testing-env/0] sandbox started',
            '[testing-env/0] ready -> acquired',
            '[testing-env/0] acquired -> setting_up',
            '[testing-env/0/session1] shell: "feature1" setup session',
            '[testing-env/0/feature1] feature setup session',
            '[testing-env/0/session1] shell: "feature2" setup session',
            '[testing-env/0/feature2] feature setup session',
            '[testing-env/0] setting_up -> in_session',
            "[testing-env/0] session 'session1' started",
            '[testing-env/0/session1] shell: echo "hello"',
            '[testing-env/0] in_session -> exiting_session',
            '[testing-env/0/session1] shell: "feature1" teardown session',
            '[testing-env/0/feature1] feature teardown session',
            '[testing-env/0/session1] shell: "feature2" teardown session',
            '[testing-env/0/feature2] feature teardown session',
            "[testing-env/0] session 'session1' ended",
            '[testing-env/0] exiting_session -> acquired',
            '[testing-env/0] acquired -> shutting_down',
            '[testing-env/0] shell: "feature1" teardown',
            '[testing-env/0/feature1] feature teardown',
            '[testing-env/0] shell: "feature2" teardown',
            '[testing-env/0/feature2] feature teardown',
            '[testing-env/0] shutting_down -> offline',
            '[testing-env/0] sandbox shutdown with SandboxStateError',
            '[testing-env] environment shutdown'
        ]
    )

  def test_feature_setup_non_state_error(self):
    env = self._create_env(
        features={
            'feature1': TestingFeature(),
            'feature2': TestingFeature(
                simulate_setup_error=ValueError
            ),
        },
    )
    with env:
      with self.assertRaises(ValueError):
        with env.sandbox('session1'):
          pass
      self.assertEqual(
          self.event_handler.logs,
          [
              '[testing-env] environment started',
              '[testing-env/0] shell: "feature1" setup',
              '[testing-env/0/feature1] feature setup',
              '[testing-env/0/feature2] feature setup with ValueError',
              '[testing-env/0] sandbox started with ValueError',
              '[testing-env/0] created -> shutting_down',
              '[testing-env/0] shell: "feature1" teardown',
              '[testing-env/0/feature1] feature teardown',
              '[testing-env/0] shell: "feature2" teardown',
              '[testing-env/0/feature2] feature teardown',
              '[testing-env/0] shutting_down -> offline',
              '[testing-env/0] sandbox shutdown'
          ]
      )

  def test_feature_setup_state_error(self):
    env = self._create_env(
        features={
            'feature1': TestingFeature(
                simulate_setup_error=interface.SandboxStateError
            ),
            'feature2': TestingFeature(),
        },
    )
    with env:
      with self.assertRaises(interface.EnvironmentOutageError):
        with env.sandbox('session1'):
          pass
      self.assertEqual(
          self.event_handler.logs,
          [
              '[testing-env] environment started',
              '[testing-env/0/feature1] feature setup with SandboxStateError',
              '[testing-env/0] sandbox started with SandboxStateError',
              '[testing-env/0] created -> shutting_down',
              '[testing-env/0] shell: "feature1" teardown',
              '[testing-env/0/feature1] feature teardown',
              '[testing-env/0] shutting_down -> offline',
              '[testing-env/0] sandbox shutdown',
              '[testing-env] environment shutdown'
          ]
      )

  def test_feature_teardown_non_state_error(self):
    env = self._create_env(
        features={
            'feature1': TestingFeature(),
            'feature2': TestingFeature(
                simulate_teardown_error=ValueError
            ),
        },
    )
    with env:
      with self.assertRaises(interface.FeatureTeardownError):
        with env.sandbox('session1'):
          pass
      self.assertEqual(
          self.event_handler.logs,
          [
              '[testing-env] environment started',
              '[testing-env/0] shell: "feature1" setup',
              '[testing-env/0/feature1] feature setup',
              '[testing-env/0] shell: "feature2" setup',
              '[testing-env/0/feature2] feature setup',
              '[testing-env/0] created -> ready',
              '[testing-env/0] sandbox started',
              '[testing-env/0] ready -> acquired',
              '[testing-env/0] acquired -> setting_up',
              '[testing-env/0/session1] shell: "feature1" setup session',
              '[testing-env/0/feature1] feature setup session',
              '[testing-env/0/session1] shell: "feature2" setup session',
              '[testing-env/0/feature2] feature setup session',
              '[testing-env/0] setting_up -> in_session',
              "[testing-env/0] session 'session1' started",
              '[testing-env/0] in_session -> exiting_session',
              '[testing-env/0/session1] shell: "feature1" teardown session',
              '[testing-env/0/feature1] feature teardown session',
              '[testing-env/0/session1] shell: "feature2" teardown session',
              '[testing-env/0/feature2] feature teardown session',
              "[testing-env/0] session 'session1' ended",
              '[testing-env/0] exiting_session -> acquired',
              '[testing-env/0] acquired -> shutting_down',
              '[testing-env/0] shell: "feature1" teardown',
              '[testing-env/0/feature1] feature teardown',
              '[testing-env/0/feature2] feature teardown with ValueError',
              '[testing-env/0] shutting_down -> offline',
              '[testing-env/0] sandbox shutdown with FeatureTeardownError',
          ]
      )

  def test_feature_teardown_state_error(self):
    env = self._create_env(
        features={
            'feature1': TestingFeature(
                simulate_teardown_error=interface.SandboxStateError
            ),
            'feature2': TestingFeature(
            ),
        },
    )
    with env:
      with env.sandbox('session1') as sb:
        pass
      self.assertEqual(len(sb.state_errors), 1)
      self.assertEqual(
          self.event_handler.logs,
          [
              '[testing-env] environment started',
              '[testing-env/0] shell: "feature1" setup',
              '[testing-env/0/feature1] feature setup',
              '[testing-env/0] shell: "feature2" setup',
              '[testing-env/0/feature2] feature setup',
              '[testing-env/0] created -> ready',
              '[testing-env/0] sandbox started',
              '[testing-env/0] ready -> acquired',
              '[testing-env/0] acquired -> setting_up',
              '[testing-env/0/session1] shell: "feature1" setup session',
              '[testing-env/0/feature1] feature setup session',
              '[testing-env/0/session1] shell: "feature2" setup session',
              '[testing-env/0/feature2] feature setup session',
              '[testing-env/0] setting_up -> in_session',
              "[testing-env/0] session 'session1' started",
              '[testing-env/0] in_session -> exiting_session',
              '[testing-env/0/session1] shell: "feature1" teardown session',
              '[testing-env/0/feature1] feature teardown session',
              '[testing-env/0/session1] shell: "feature2" teardown session',
              '[testing-env/0/feature2] feature teardown session',
              "[testing-env/0] session 'session1' ended",
              '[testing-env/0] exiting_session -> acquired',
              '[testing-env/0] acquired -> shutting_down',
              '[testing-env/0/feature1] feature teardown with SandboxStateError',  # pylint: disable=line-too-long
              '[testing-env/0] shell: "feature2" teardown',
              '[testing-env/0/feature2] feature teardown',
              '[testing-env/0] shutting_down -> offline',
              '[testing-env/0] sandbox shutdown with FeatureTeardownError'
          ]
      )

  def test_feature_setup_session_non_state_error(self):
    env = self._create_env(
        features={
            'feature1': TestingFeature(),
            'feature2': TestingFeature(
                simulate_setup_session_error=ValueError
            ),
        },
    )
    with env:
      with self.assertRaises(ValueError):
        with env.sandbox('session1') as sb:
          sb.shell('echo "hello"')
      self.assertEqual(
          self.event_handler.logs,
          [
              '[testing-env] environment started',
              '[testing-env/0] shell: "feature1" setup',
              '[testing-env/0/feature1] feature setup',
              '[testing-env/0] shell: "feature2" setup',
              '[testing-env/0/feature2] feature setup',
              '[testing-env/0] created -> ready',
              '[testing-env/0] sandbox started',
              '[testing-env/0] ready -> acquired',
              '[testing-env/0] acquired -> setting_up',
              '[testing-env/0/session1] shell: "feature1" setup session',
              '[testing-env/0/feature1] feature setup session',
              '[testing-env/0/feature2] feature setup session with ValueError',
              "[testing-env/0] session 'session1' started with ValueError",
              '[testing-env/0] setting_up -> shutting_down',
              '[testing-env/0/session1] shell: "feature1" teardown',
              '[testing-env/0/feature1] feature teardown',
              '[testing-env/0/session1] shell: "feature2" teardown',
              '[testing-env/0/feature2] feature teardown',
              '[testing-env/0] shutting_down -> offline',
              '[testing-env/0] sandbox shutdown'
          ]
      )

  def test_feature_teardown_session_non_state_error(self):
    env = self._create_env(
        features={
            'feature1': TestingFeature(
                simulate_teardown_session_error=ValueError
            ),
            'feature2': TestingFeature(),
        },
    )
    with env:
      with self.assertRaises(interface.SessionTeardownError):
        with env.sandbox('session1') as sb:
          sb.shell('echo "hello"')
      self.assertEqual(sb.status, interface.Sandbox.Status.OFFLINE)
      self.assertEqual(
          self.event_handler.logs,
          [
              '[testing-env] environment started',
              '[testing-env/0] shell: "feature1" setup',
              '[testing-env/0/feature1] feature setup',
              '[testing-env/0] shell: "feature2" setup',
              '[testing-env/0/feature2] feature setup',
              '[testing-env/0] created -> ready',
              '[testing-env/0] sandbox started',
              '[testing-env/0] ready -> acquired',
              '[testing-env/0] acquired -> setting_up',
              '[testing-env/0/session1] shell: "feature1" setup session',
              '[testing-env/0/feature1] feature setup session',
              '[testing-env/0/session1] shell: "feature2" setup session',
              '[testing-env/0/feature2] feature setup session',
              '[testing-env/0] setting_up -> in_session',
              "[testing-env/0] session 'session1' started",
              '[testing-env/0/session1] shell: echo "hello"',
              '[testing-env/0] in_session -> exiting_session',
              '[testing-env/0/feature1] feature teardown session with ValueError',  # pylint: disable=line-too-long
              '[testing-env/0/session1] shell: "feature2" teardown session',
              '[testing-env/0/feature2] feature teardown session',
              "[testing-env/0] session 'session1' ended",
              '[testing-env/0] exiting_session -> acquired',
              '[testing-env/0] acquired -> shutting_down',
              '[testing-env/0] shell: "feature1" teardown',
              '[testing-env/0/feature1] feature teardown',
              '[testing-env/0] shell: "feature2" teardown',
              '[testing-env/0/feature2] feature teardown',
              '[testing-env/0] shutting_down -> offline',
              '[testing-env/0] sandbox shutdown'
          ]
      )

  def test_feature_teardown_session_state_error(self):
    env = self._create_env(
        features={
            'feature1': TestingFeature(
                simulate_teardown_session_error=interface.SandboxStateError
            ),
            'feature2': TestingFeature(),
        },
    )
    with env:
      with env.sandbox('session1') as sb:
        sb.shell('echo "hello"')
      self.assertEqual(len(sb.state_errors), 1)
      self.assertEqual(sb.status, interface.Sandbox.Status.OFFLINE)
      self.assertEqual(
          self.event_handler.logs,
          [
              '[testing-env] environment started',
              '[testing-env/0] shell: "feature1" setup',
              '[testing-env/0/feature1] feature setup',
              '[testing-env/0] shell: "feature2" setup',
              '[testing-env/0/feature2] feature setup',
              '[testing-env/0] created -> ready',
              '[testing-env/0] sandbox started',
              '[testing-env/0] ready -> acquired',
              '[testing-env/0] acquired -> setting_up',
              '[testing-env/0/session1] shell: "feature1" setup session',
              '[testing-env/0/feature1] feature setup session',
              '[testing-env/0/session1] shell: "feature2" setup session',
              '[testing-env/0/feature2] feature setup session',
              '[testing-env/0] setting_up -> in_session',
              "[testing-env/0] session 'session1' started",
              '[testing-env/0/session1] shell: echo "hello"',
              '[testing-env/0] in_session -> exiting_session',
              '[testing-env/0/feature1] feature teardown session with SandboxStateError',  # pylint: disable=line-too-long
              '[testing-env/0/session1] shell: "feature2" teardown session',
              '[testing-env/0/feature2] feature teardown session',
              "[testing-env/0] session 'session1' ended with SandboxStateError",
              '[testing-env/0] exiting_session -> acquired',
              '[testing-env/0] acquired -> shutting_down',
              '[testing-env/0] shell: "feature1" teardown',
              '[testing-env/0/feature1] feature teardown',
              '[testing-env/0] shell: "feature2" teardown',
              '[testing-env/0/feature2] feature teardown',
              '[testing-env/0] shutting_down -> offline',
              '[testing-env/0] sandbox shutdown'
          ]
      )

  def test_feature_teardown_session_calling_end_session(self):
    env = self._create_env(
        features={
            'feature1': TestingFeature(
                call_end_session_on_teardown_session=True
            ),
            'feature2': TestingFeature(),
        },
    )
    with env:
      with env.sandbox('session1') as sb:
        sb.shell('echo "hello"')
      self.assertEqual(
          self.event_handler.logs,
          [
              '[testing-env] environment started',
              '[testing-env/0] shell: "feature1" setup',
              '[testing-env/0/feature1] feature setup',
              '[testing-env/0] shell: "feature2" setup',
              '[testing-env/0/feature2] feature setup',
              '[testing-env/0] created -> ready',
              '[testing-env/0] sandbox started',
              '[testing-env/0] ready -> acquired',
              '[testing-env/0] acquired -> setting_up',
              '[testing-env/0/session1] shell: "feature1" setup session',
              '[testing-env/0/feature1] feature setup session',
              '[testing-env/0/session1] shell: "feature2" setup session',
              '[testing-env/0/feature2] feature setup session',
              '[testing-env/0] setting_up -> in_session',
              "[testing-env/0] session 'session1' started",
              '[testing-env/0/session1] shell: echo "hello"',
              '[testing-env/0] in_session -> exiting_session',
              '[testing-env/0/session1] shell: "feature1" teardown session',
              '[testing-env/0/feature1] feature teardown session',
              '[testing-env/0/session1] shell: "feature2" teardown session',
              '[testing-env/0/feature2] feature teardown session',
              "[testing-env/0] session 'session1' ended",
              '[testing-env/0] exiting_session -> acquired',
              '[testing-env/0] acquired -> shutting_down',
              '[testing-env/0] shell: "feature1" teardown',
              '[testing-env/0/feature1] feature teardown',
              '[testing-env/0] shell: "feature2" teardown',
              '[testing-env/0/feature2] feature teardown',
              '[testing-env/0] shutting_down -> offline',
              '[testing-env/0] sandbox shutdown'
          ]
      )

  def test_session_activity_non_state_error(self):
    env = self._create_env(
        pool_size=1,
        features={
            'feature1': TestingFeature(),
        },
    )
    with env:
      with env.sandbox('session1') as sb:
        with self.assertRaises(ValueError):
          sb.shell('echo foo', raise_error=ValueError)
        self.assertEqual(len(sb.state_errors), 0)
        sb.shell('echo bar')
        self.assertEqual(sb.status, interface.Sandbox.Status.IN_SESSION)
      sb.wait_until_not(interface.Sandbox.Status.SETTING_UP)
      self.assertEqual(sb.status, interface.Sandbox.Status.READY)
      self.assertEqual(
          self.event_handler.logs,
          [
              '[testing-env/0:0] shell: "feature1" setup',
              '[testing-env/0:0/feature1] feature setup',
              '[testing-env/0:0] shell: "feature1" setup session',
              '[testing-env/0:0/feature1] feature setup session',
              '[testing-env/0:0] created -> ready',
              '[testing-env/0:0] sandbox started',
              '[testing-env] environment started',
              '[testing-env/0:0] ready -> acquired',
              '[testing-env/0:0] acquired -> setting_up',
              '[testing-env/0:0] setting_up -> in_session',
              "[testing-env/0:0] session 'session1' started",
              '[testing-env/0:0/session1] shell: echo foo with ValueError',
              '[testing-env/0:0/session1] shell: echo bar',
              '[testing-env/0:0] in_session -> exiting_session',
              '[testing-env/0:0/session1] shell: "feature1" teardown session',
              '[testing-env/0:0/feature1] feature teardown session',
              "[testing-env/0:0] session 'session1' ended",
              '[testing-env/0:0] exiting_session -> setting_up',
              '[testing-env/0:0] shell: "feature1" setup session',
              '[testing-env/0:0/feature1] feature setup session',
              '[testing-env/0:0] setting_up -> ready',
          ]
      )

  def test_session_activity_state_error(self):
    env = self._create_env(
        pool_size=1,
        features={
            'feature1': TestingFeature(),
        },
    )
    with env:
      with env.sandbox('session1') as sb:
        with self.assertRaises(interface.SandboxStateError):
          sb.shell('echo foo', raise_error=RuntimeError)
        self.assertEqual(len(sb.state_errors), 1)
        self.assertEqual(sb.status, interface.Sandbox.Status.OFFLINE)
        sb.shell('echo bar')
      self.assertEqual(
          self.event_handler.logs,
          [
              # pylint: disable=line-too-long
              '[testing-env/0:0] shell: "feature1" setup',
              '[testing-env/0:0/feature1] feature setup',
              '[testing-env/0:0] shell: "feature1" setup session',
              '[testing-env/0:0/feature1] feature setup session',
              '[testing-env/0:0] created -> ready',
              '[testing-env/0:0] sandbox started',
              '[testing-env] environment started',
              '[testing-env/0:0] ready -> acquired',
              '[testing-env/0:0] acquired -> setting_up',
              '[testing-env/0:0] setting_up -> in_session',
              "[testing-env/0:0] session 'session1' started",
              '[testing-env/0:0/session1] shell: echo foo with RuntimeError',
              '[testing-env/0:0] in_session -> exiting_session',
              '[testing-env/0:0/session1] shell: "feature1" teardown session',
              '[testing-env/0:0/feature1] feature teardown session',
              "[testing-env/0:0] session 'session1' ended with SandboxStateError",
              '[testing-env/0:0] exiting_session -> acquired',
              '[testing-env/0:0] acquired -> shutting_down',
              '[testing-env/0:0] shell: "feature1" teardown',
              '[testing-env/0:0/feature1] feature teardown',
              '[testing-env/0:0] shutting_down -> offline',
              '[testing-env/0:0] sandbox shutdown',
              '[testing-env/0:0] shell: echo bar',
              # pylint: enable=line-too-long
          ]
      )


class SandboxActivityTests(unittest.TestCase):

  def test_session_id(self):
    env = TestingEnvironment(
        features={'test_feature': TestingFeature()},
        pool_size=0
    )
    with env:
      with env.sandbox() as sb:
        self.assertRegex(sb.session_id, r'session-[0-9a-f]{7}')

      self.assertEqual(
          env.test_feature.show_session_id(session_id='session1'),
          'session1'
      )
      self.assertRegex(
          env.test_feature.show_session_id(),
          r'session-[0-9a-f]{7}'
      )

    with self.assertRaisesRegex(ValueError, '`session_id` should not be used'):
      @base_sandbox.sandbox_service()
      def foo(session_id: str):
        del session_id

  def test_ping_error(self):
    env = TestingEnvironment(
        features={'test_feature': TestingFeature(housekeep_interval=0)},
        pool_size=1,
        sandbox_keepalive_interval=0,
    )
    with env:
      with env.sandbox('session1') as sb:
        sb.rebind(
            simulate_ping_error=interface.SandboxStateError,
            skip_notification=True
        )
        sb.wait_until_next_housekeep()
        self.assertEqual(sb.status, sb.Status.OFFLINE)

  def test_housekeep_error(self):
    event_handler = TestingEventHandler(log_housekeep=False)
    env = TestingEnvironment(
        features={'test_feature': TestingFeature(housekeep_interval=0)},
        pool_size=1,
        housekeep_interval=1.0,
        outage_grace_period=0,
        outage_retry_interval=0.1,
        sandbox_keepalive_interval=0,
        event_handlers=[event_handler],
    )
    with env:
      with env.sandbox('session1') as sb:
        self.assertEqual(len(env.sandbox_pool), 1)
        self.assertEqual(sb.status, interface.Sandbox.Status.IN_SESSION)
        self.assertEqual(sb.session_id, 'session1')
        housekeep_count = sb.housekeep_counter
        sb.test_feature.rebind(
            simulate_housekeep_error=interface.SandboxStateError,
            skip_notification=True
        )
        while sb.housekeep_counter == housekeep_count or (
            sb.status == interface.Sandbox.Status.IN_SESSION
        ):
          time.sleep(0.01)
        self.assertEqual(sb.status, interface.Sandbox.Status.OFFLINE)
      env.wait_for_housekeeping()
    self.assertEqual(
        event_handler.logs,
        [
            '[testing-env/0:0] shell: "test_feature" setup',
            '[testing-env/0:0/test_feature] feature setup',
            '[testing-env/0:0] shell: "test_feature" setup session',
            '[testing-env/0:0] sandbox started',
            '[testing-env] environment started',
            "[testing-env/0:0] session 'session1' started",
            '[testing-env/0:0/session1] shell: "test_feature" teardown session',
            "[testing-env/0:0] session 'session1' ended with SandboxStateError",
            '[testing-env/0:0] shell: "test_feature" teardown',
            '[testing-env/0:0/test_feature] feature teardown',
            '[testing-env/0:0] sandbox shutdown',
            '[testing-env/0:1] shell: "test_feature" setup',
            '[testing-env/0:1/test_feature] feature setup',
            '[testing-env/0:1] shell: "test_feature" setup session',
            '[testing-env/0:1] sandbox started',
            '[testing-env/0:1] shell: "test_feature" teardown',
            '[testing-env/0:1/test_feature] feature teardown',
            '[testing-env/0:1] sandbox shutdown',
            '[testing-env] environment shutdown'
        ]
    )

  def test_remove_event_handler(self):
    env = TestingEnvironment(
        features={'test_feature': TestingFeature(housekeep_interval=0)},
        pool_size=1,
        outage_grace_period=0,
        outage_retry_interval=0,
        sandbox_keepalive_interval=0,
    )
    event_handler = TestingEventHandler()
    with env:
      with env.sandbox('session1') as sb:
        sb.add_event_handler(event_handler)
        sb.shell('test_feature')
        sb.remove_event_handler(event_handler)
      events = list(event_handler.logs)
      sb.wait_until_not(interface.Sandbox.Status.SETTING_UP)
      self.assertGreater(len(events), 0)
      with env.sandbox('session2') as sb:
        sb.shell('test_feature')
      self.assertEqual(len(events), len(event_handler.logs))


class SandboxServiceTests(unittest.TestCase):

  def setUp(self):
    super().setUp()
    self.maxDiff = None
    self.event_handler = TestingEventHandler()
    self.env = TestingEnvironment(
        features={'test_feature': TestingFeature()},
        pool_size=0,
        outage_grace_period=0,
        outage_retry_interval=0,
        sandbox_keepalive_interval=0,
        event_handlers=[self.event_handler],
        random_seed=1,
    )

  def test_service_call_activity_log(self):

    class CustomEventHandler(event_handler_base.EventHandler):

      def __init__(self):
        self.calls = []

      def on_sandbox_activity(
          self,
          name: str,
          environment: interface.Environment,
          sandbox: interface.Sandbox,
          feature: interface.Feature | None,
          session_id: str | None,
          duration: float,
          error: BaseException | None,
          **kwargs: Any):
        self.calls.append((session_id, name, kwargs))

    event_handler = CustomEventHandler()
    env = TestingEnvironment(
        features={'test_feature': TestingFeature()},
        pool_size=0,
        event_handlers=[event_handler],
    )
    with env:
      env.test_feature.call_with_varargs(
          'sum', 1, 2, debug=True, session_id='session1'
      )
    self.assertEqual(
        event_handler.calls,
        [
            (None, 'shell', {'code': '"test_feature" setup'}),
            ('session1', 'shell', {'code': '"test_feature" setup session'}),
            ('session1', 'test_feature.call_with_varargs', {'args': (1, 2), 'code': 'sum', 'debug': True}),   # pylint: disable=line-too-long
            ('session1', 'shell', {'code': '"test_feature" teardown session'}),
            (None, 'shell', {'code': '"test_feature" teardown'}),
        ]
    )

  def test_service_call_from_feature(self):
    with self.env:
      with self.env.sandbox('session1') as sb:
        self.assertEqual(sb.test_feature.num_shell_calls(), 2)
        self.assertEqual(sb.test_feature.num_shell_calls(), 2)
    self.assertEqual(
        self.event_handler.logs,
        [
            # pylint: disable=line-too-long
            '[testing-env] environment started',
            '[testing-env/0] shell: "test_feature" setup',
            '[testing-env/0/test_feature] feature setup',
            '[testing-env/0] sandbox started',
            '[testing-env/0/session1] shell: "test_feature" setup session',
            "[testing-env/0] session 'session1' started",
            '[testing-env/0/session1/test_feature] test_feature.num_shell_calls: None',
            '[testing-env/0/session1/test_feature] test_feature.num_shell_calls: None',
            '[testing-env/0/session1] shell: "test_feature" teardown session',
            "[testing-env/0] session 'session1' ended",
            '[testing-env/0] shell: "test_feature" teardown',
            '[testing-env/0/test_feature] feature teardown',
            '[testing-env/0] sandbox shutdown',
            '[testing-env] environment shutdown'
            # pylint: enable=line-too-long
        ]
    )

  def test_service_call_from_feature_with_error(self):
    with self.env:
      with self.env.sandbox('session1') as sb:
        with self.assertRaises(interface.SandboxStateError):
          sb.test_feature.bad_shell_call()
        self.assertEqual(sb.status, interface.Sandbox.Status.OFFLINE)

    self.assertEqual(
        self.event_handler.logs,
        [
            # pylint: disable=line-too-long
            '[testing-env] environment started',
            '[testing-env/0] shell: "test_feature" setup',
            '[testing-env/0/test_feature] feature setup',
            '[testing-env/0] sandbox started',
            '[testing-env/0/session1] shell: "test_feature" setup session',
            "[testing-env/0] session 'session1' started",
            '[testing-env/0/session1] shell: bad command with RuntimeError',
            '[testing-env/0/session1/test_feature] test_feature.bad_shell_call: None with SandboxStateError',
            '[testing-env/0/session1] shell: "test_feature" teardown session',
            "[testing-env/0] session 'session1' ended with SandboxStateError",
            '[testing-env/0] shell: "test_feature" teardown',
            '[testing-env/0/test_feature] feature teardown',
            '[testing-env/0] sandbox shutdown',
            '[testing-env] environment shutdown'
            # pylint: enable=line-too-long
        ]
    )

  def test_service_call_from_environment(self):
    with self.env:
      self.assertEqual(self.env.test_feature.num_shell_calls(), 2)
    self.assertEqual(
        self.event_handler.logs,
        [
            # pylint: disable=line-too-long
            '[testing-env] environment started',
            '[testing-env/0] shell: "test_feature" setup',
            '[testing-env/0/test_feature] feature setup',
            '[testing-env/0] sandbox started',
            '[testing-env/0/session-2291d8c] shell: "test_feature" setup session',
            "[testing-env/0] session 'session-2291d8c' started",
            '[testing-env/0/session-2291d8c/test_feature] test_feature.num_shell_calls: None',
            '[testing-env/0/session-2291d8c] shell: "test_feature" teardown session',
            "[testing-env/0] session 'session-2291d8c' ended",
            '[testing-env/0] shell: "test_feature" teardown',
            '[testing-env/0/test_feature] feature teardown',
            '[testing-env/0] sandbox shutdown',
            '[testing-env] environment shutdown'
            # pylint: enable=line-too-long
        ]
    )

  def test_service_call_from_environment_with_error(self):
    with self.env:
      with self.assertRaises(interface.SandboxStateError):
        self.env.test_feature.bad_shell_call(session_id='session1')
    self.assertEqual(
        self.event_handler.logs,
        [
            # pylint: disable=line-too-long
            '[testing-env] environment started',
            '[testing-env/0] shell: "test_feature" setup',
            '[testing-env/0/test_feature] feature setup',
            '[testing-env/0] sandbox started',
            '[testing-env/0/session1] shell: "test_feature" setup session',
            "[testing-env/0] session 'session1' started",
            '[testing-env/0/session1] shell: bad command with RuntimeError',
            '[testing-env/0/session1/test_feature] test_feature.bad_shell_call: None with SandboxStateError',
            '[testing-env/0/session1] shell: "test_feature" teardown session',
            "[testing-env/0] session 'session1' ended with SandboxStateError",
            '[testing-env/0] shell: "test_feature" teardown',
            '[testing-env/0/test_feature] feature teardown',
            '[testing-env/0] sandbox shutdown',
            '[testing-env] environment shutdown'
            # pylint: enable=line-too-long
        ]
    )

  def test_service_context_manager_from_feature(self):
    with self.env:
      with self.env.sandbox('session1') as sb:
        with sb.test_feature.my_service() as service:
          service.do('hello')
        sb.shell('foo')
        self.assertEqual(sb.status, interface.Sandbox.Status.IN_SESSION)
    self.assertEqual(
        self.event_handler.logs,
        [
            # pylint: disable=line-too-long
            '[testing-env] environment started',
            '[testing-env/0] shell: "test_feature" setup',
            '[testing-env/0/test_feature] feature setup',
            '[testing-env/0] sandbox started',
            '[testing-env/0/session1] shell: "test_feature" setup session',
            "[testing-env/0] session 'session1' started",
            '[testing-env/0/session1] shell: hello',
            '[testing-env/0/session1/test_feature] test_feature.my_service: None',
            '[testing-env/0/session1] shell: foo',
            '[testing-env/0/session1] shell: "test_feature" teardown session',
            "[testing-env/0] session 'session1' ended",
            '[testing-env/0] shell: "test_feature" teardown',
            '[testing-env/0/test_feature] feature teardown',
            '[testing-env/0] sandbox shutdown',
            '[testing-env] environment shutdown'
            # pylint: enable=line-too-long
        ]
    )

  def test_service_context_manager_from_feature_with_error(self):
    with self.env:
      with self.env.sandbox('session1') as sb:
        with self.assertRaises(interface.SandboxStateError):
          with sb.test_feature.my_service() as service:
            service.do('hello', raise_error=interface.SandboxStateError)
        self.assertEqual(sb.status, interface.Sandbox.Status.OFFLINE)
    self.assertEqual(
        self.event_handler.logs,
        [
            # pylint: disable=line-too-long
            '[testing-env] environment started',
            '[testing-env/0] shell: "test_feature" setup',
            '[testing-env/0/test_feature] feature setup',
            '[testing-env/0] sandbox started',
            '[testing-env/0/session1] shell: "test_feature" setup session',
            "[testing-env/0] session 'session1' started",
            '[testing-env/0/session1] shell: hello with SandboxStateError',
            '[testing-env/0/session1/test_feature] test_feature.my_service: None with SandboxStateError',
            '[testing-env/0/session1] shell: "test_feature" teardown session',
            "[testing-env/0] session 'session1' ended with SandboxStateError",
            '[testing-env/0] shell: "test_feature" teardown',
            '[testing-env/0/test_feature] feature teardown',
            '[testing-env/0] sandbox shutdown',
            '[testing-env] environment shutdown'
            # pylint: enable=line-too-long
        ]
    )

  def test_service_context_manager_from_environment(self):
    with self.env:
      with self.env.test_feature.my_service(session_id='session1') as service:
        service.do('foo')
      with self.env.test_feature.my_service() as service:
        service.do('bar')
    self.assertEqual(
        self.event_handler.logs,
        [
            # pylint: disable=line-too-long
            '[testing-env] environment started',
            '[testing-env/0] shell: "test_feature" setup',
            '[testing-env/0/test_feature] feature setup',
            '[testing-env/0] sandbox started',
            '[testing-env/0/session1] shell: "test_feature" setup session',
            "[testing-env/0] session 'session1' started",
            '[testing-env/0/session1] shell: foo',
            '[testing-env/0/session1/test_feature] test_feature.my_service: None',
            '[testing-env/0/session1] shell: "test_feature" teardown session',
            "[testing-env/0] session 'session1' ended",
            '[testing-env/0] shell: "test_feature" teardown',
            '[testing-env/0/test_feature] feature teardown',
            '[testing-env/0] sandbox shutdown',
            '[testing-env/1] shell: "test_feature" setup',
            '[testing-env/1/test_feature] feature setup',
            '[testing-env/1] sandbox started',
            '[testing-env/1/session-2291d8c] shell: "test_feature" setup session',
            "[testing-env/1] session 'session-2291d8c' started",
            '[testing-env/1/session-2291d8c] shell: bar',
            '[testing-env/1/session-2291d8c/test_feature] test_feature.my_service: None',
            '[testing-env/1/session-2291d8c] shell: "test_feature" teardown session',
            "[testing-env/1] session 'session-2291d8c' ended",
            '[testing-env/1] shell: "test_feature" teardown',
            '[testing-env/1/test_feature] feature teardown',
            '[testing-env/1] sandbox shutdown',
            '[testing-env] environment shutdown'
            # pylint: enable=line-too-long
        ]
    )

  def test_service_context_manager_from_environment_with_error(self):
    with self.env:
      with self.assertRaises(interface.SandboxStateError):
        with self.env.test_feature.my_service() as service:
          service.do('hello', raise_error=interface.SandboxStateError)
    self.assertEqual(
        self.event_handler.logs,
        [
            # pylint: disable=line-too-long
            '[testing-env] environment started',
            '[testing-env/0] shell: "test_feature" setup',
            '[testing-env/0/test_feature] feature setup',
            '[testing-env/0] sandbox started',
            '[testing-env/0/session-2291d8c] shell: "test_feature" setup session',
            "[testing-env/0] session 'session-2291d8c' started",
            '[testing-env/0/session-2291d8c] shell: hello with SandboxStateError',
            '[testing-env/0/session-2291d8c/test_feature] test_feature.my_service: None with SandboxStateError',
            '[testing-env/0/session-2291d8c] shell: "test_feature" teardown session',
            "[testing-env/0] session 'session-2291d8c' ended with SandboxStateError",
            '[testing-env/0] shell: "test_feature" teardown',
            '[testing-env/0/test_feature] feature teardown',
            '[testing-env/0] sandbox shutdown',
            '[testing-env] environment shutdown'
            # pylint: enable=line-too-long
        ]
    )


if __name__ == '__main__':
  unittest.main()



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common base class for sandbox-based features.

This module provides an base class `BaseFeature` for sandbox-based features,
which provides event handlers for the feature lifecycle events, which can be
overridden by subclasses to provide custom behaviors. Please note that this base
class is intended to provide a convenient way to implement features, and not
all feature implementations need to subclass it. Also `BaseFeature` is not
coupled with `BaseEnvironment` and `BaseSandbox`, and is expected to work with
the `Environment` and `Sandbox` interfaces directly.
"""

import functools
import os
import time
from typing import Annotated, Callable

from langfun.env import interface
import pyglove as pg


class BaseFeature(interface.Feature):
  """Common base class for sandbox-based features."""

  housekeep_interval: Annotated[
      float | None,
      'Interval in seconds for feature housekeeping.'
  ] = None

  #
  # Subclasses can override:
  #

  def _setup(self) -> None:
    """Subclasses can override this for custom setup.

    NOTE: always call super()._setup() at the beginning of the implementation.
    """

  def _teardown(self) -> None:
    """Subclasses can override this for custom teardown.

    NOTE: always call super()._teardown() at the end of the implementation.
    """

  def _setup_session(self) -> None:
    """Subclasses can override this for custom setup session.

    NOTE: always call super()._setup_session() at the beginning of the
    implementation.
    """

  def _teardown_session(self) -> None:
    """Subclasses can override this for custom teardown session.

    NOTE: always call super()._teardown_session() at the end of the
    implementation.
    """

  def _housekeep(self) -> None:
    """Performs housekeeping for the feature.

    NOTE: always call super()._housekeep() at the beginning of the
    implementation.
    """

  #
  # Init and properties
  #

  def _on_bound(self) -> None:
    """Called when the feature is bound."""
    super()._on_bound()
    self._sandbox = None
    self._housekeep_counter = 0

  @functools.cached_property
  def name(self) -> str:
    """Returns the name of the feature."""
    assert isinstance(self.sym_parent, dict), 'Feature is not put into a dict.'
    return self.sym_path.key

  def _on_parent_change(
      self,
      old_parent: pg.Symbolic | None,
      new_parent: pg.Symbolic | None
  ) -> None:
    """Called when the feature is bound."""
    super()._on_parent_change(old_parent, new_parent)
    self.__dict__.pop('name', None)

  @property
  def sandbox(self) -> interface.Sandbox:
    """Returns the sandbox that the feature is running in."""
    assert self._sandbox is not None, 'Feature has not been set up yet.'
    return self._sandbox

  @property
  def working_dir(self) -> str | None:
    """Returns the working directory of the feature."""
    sandbox_workdir = self.sandbox.working_dir
    if sandbox_workdir is None:
      return None
    return os.path.join(sandbox_workdir, self.name)

  #
  # Setup and teardown of the feature.
  #

  def _do(
      self,
      action: Callable[[], None],
      event_handler: Callable[..., None],
  ) -> None:
    """Triggers an event handler."""
    error = None
    start_time = time.time()
    try:
      action()
    except BaseException as e:  # pylint: disable=broad-except
      error = e
      raise
    finally:
      event_handler(duration=time.time() - start_time, error=error)

  def setup(self, sandbox: interface.Sandbox) -> None:
    """Sets up the feature."""
    self._sandbox = sandbox
    self._do(self._setup, self.on_setup)

  def teardown(self) -> None:
    """Tears down the feature."""
    self._do(self._teardown, event_handler=self.on_teardown)

  def setup_session(self) -> None:
    """Sets up the feature for a user session."""
    self._do(self._setup_session, event_handler=self.on_setup_session)

  def teardown_session(self) -> None:
    """Teardowns the feature for a user session."""
    self._do(self._teardown_session, self.on_teardown_session)

  #
  # Housekeeping.
  #

  def housekeep(self) -> None:
    """Performs housekeeping for the feature."""
    try:
      self._do(self._housekeep, self.on_housekeep)
    finally:
      self._housekeep_counter += 1

  #
  # Event handlers subclasses can override.
  #

  def on_setup(
      self,
      duration: float,
      error: BaseException | None = None
  ) -> None:
    """Called when the feature is setup."""
    self.sandbox.on_feature_setup(self, duration, error)

  def on_teardown(
      self,
      duration: float,
      error: BaseException | None = None
  ) -> None:
    """Called when the feature is teardown."""
    self.sandbox.on_feature_teardown(self, duration, error)

  def on_housekeep(
      self,
      duration: float,
      error: BaseException | None = None,
      **kwargs
  ) -> None:
    """Called when the feature has done housekeeping."""
    self.sandbox.on_feature_housekeep(
        self, self._housekeep_counter, duration, error, **kwargs
    )

  def on_setup_session(
      self,
      duration: float,
      error: BaseException | None = None,
  ) -> None:
    """Called when the feature is setup for a user session."""
    self.sandbox.on_feature_setup_session(self, duration, error)

  def on_teardown_session(
      self,
      duration: float,
      error: BaseException | None = None,
  ) -> None:
    """Called when the feature is teardown for a user session."""
    self.sandbox.on_feature_teardown_session(self, duration, error)

  def on_activity(
      self,
      name: str,
      duration: float,
      error: BaseException | None = None,
      **kwargs
  ) -> None:
    """Called when a sandbox activity is performed."""
    self.sandbox.on_activity(
        name=f'{self.name}.{name}',
        feature=self,
        error=error,
        duration=duration,
        **kwargs
    )



"""Environment event handlers."""

# pylint: disable=g-importing-member
# pylint: disable=g-bad-import-order

from langfun.env.event_handlers.base import EventHandler

from langfun.env.event_handlers.event_logger import EventLogger
from langfun.env.event_handlers.event_logger import ConsoleEventLogger

from langfun.env.event_handlers.metric_writer import MetricWriter

# pylint: enable=g-importing-member
# pylint: enable=g-bad-import-order



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""MCP session."""

import contextlib
from typing import Any, Type
import anyio
from langfun.core import async_support
from langfun.core.mcp import tool as mcp_tool
import mcp
from mcp.client import sse
from mcp.client import streamable_http
from mcp.server import fastmcp as fastmcp_lib
from mcp.shared import memory


class McpSession:
  """Langfun's MCP session.

  Compared to the standard mcp.ClientSession, Langfun's MCP session could be
  used both synchronously and asynchronously.
  """

  def __init__(self, stream) -> None:
    self._stream = stream
    self._session = None
    self._session_exit_stack = None
    self._in_session = False

    # For supporting sync context manager.
    self._sync_context_manager_exit_stack = None

  def __enter__(self) -> 'McpSession':
    exit_stack = contextlib.ExitStack()
    exit_stack.enter_context(async_support.sync_context_manager(self))
    self._sync_context_manager_exit_stack = exit_stack
    return self

  def __exit__(self, exc_type, exc_val, exc_tb) -> None:
    assert self._sync_context_manager_exit_stack is not None
    self._sync_context_manager_exit_stack.close()

  async def __aenter__(self) -> 'McpSession':
    assert self._session_exit_stack is None, 'Session cannot be re-entered.'

    self._session_exit_stack = contextlib.AsyncExitStack()
    stream_output = await self._session_exit_stack.enter_async_context(
        self._stream
    )
    assert isinstance(stream_output, tuple) and len(stream_output) in (2, 3)
    read, write = stream_output[:2]
    self._session = mcp.ClientSession(read, write)
    await self._session_exit_stack.enter_async_context(self._session)
    await self._session.initialize()
    return self

  async def __aexit__(self, exc_type, exc_val, exc_tb) -> None:
    del exc_type, exc_val, exc_tb
    if self._session is None:
      return
    assert self._session_exit_stack is not None
    await self._session_exit_stack.aclose()
    self._session = None

  def list_tools(self) -> dict[str, Type[mcp_tool.McpTool]]:
    """Lists all MCP tools synchronously."""
    return async_support.invoke_sync(self.alist_tools)

  async def alist_tools(self) -> dict[str, Type[mcp_tool.McpTool]]:
    """Lists all MCP tools asynchronously."""
    assert self._session is not None, 'MCP session is not entered.'
    return {
        t.name: mcp_tool.McpTool.make_class(t)
        for t in (await self._session.list_tools()).tools
    }

  def call_tool(
      self,
      tool: mcp_tool.McpTool,
      *,
      returns_call_result: bool = False
  ) -> Any:
    """Calls a MCP tool synchronously."""
    return async_support.invoke_sync(
        self.acall_tool,
        tool,
        returns_call_result=returns_call_result
    )

  async def acall_tool(
      self,
      tool: mcp_tool.McpTool,
      *,
      returns_call_result: bool = False
  ) -> Any:
    """Calls a MCP tool asynchronously."""
    assert self._session is not None, 'MCP session is not entered.'
    tool_call_result = await self._session.call_tool(
        tool.TOOL_NAME, tool.input_parameters()
    )
    if returns_call_result:
      return tool_call_result
    if (
        tool_call_result.structuredContent
        and 'result' in tool_call_result.structuredContent
    ):
      return tool_call_result.structuredContent['result']
    return tool_call_result.content

  @classmethod
  def from_command(
      cls,
      command: str,
      args: list[str] | None = None
  ) -> 'McpSession':
    """Creates a MCP session from a command."""
    return cls(
        mcp.stdio_client(
            mcp.StdioServerParameters(command=command, args=args or [])
        )
    )

  @classmethod
  def from_url(
      cls,
      url: str,
      headers: dict[str, str] | None = None
  ) -> 'McpSession':
    """Creates a MCP session from a URL."""
    transport = url.removesuffix('/').split('/')[-1].lower()
    if transport == 'mcp':
      return cls(streamable_http.streamablehttp_client(url, headers or {}))
    elif transport == 'sse':
      return cls(sse.sse_client(url, headers or {}))
    else:
      raise ValueError(f'Unsupported transport: {transport}')

  @classmethod
  def from_fastmcp(
      cls,
      fastmcp: fastmcp_lib.FastMCP
  ):
    return cls(_client_streams_from_fastmcp(fastmcp))


@contextlib.asynccontextmanager
async def _client_streams_from_fastmcp(fastmcp: fastmcp_lib.FastMCP):
  """Creates client streams from a MCP server."""
  server = fastmcp._mcp_server  # pylint: disable=protected-access
  async with memory.create_client_server_memory_streams(
      ) as (client_streams, server_streams):
    client_read, client_write = client_streams
    server_read, server_write = server_streams

    # Create a cancel scope for the server task
    async with anyio.create_task_group() as tg:
      tg.start_soon(
          lambda: server.run(
              server_read,
              server_write,
              server.create_initialization_options(),
              raise_exceptions=True,
          )
      )
      yield client_read, client_write



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""MCP tool."""

from typing import Annotated, Any, ClassVar

from langfun.core.structured import schema as lf_schema
import mcp
import pyglove as pg


class _McpToolMeta(pg.symbolic.ObjectMeta):

  def __repr__(self) -> str:
    return f'<tool-class \'{self.__name__}\'>'


class McpTool(pg.Object, metaclass=_McpToolMeta):
  """Base class for MCP tools."""

  TOOL_NAME: Annotated[
      ClassVar[str],
      'Tool name.'
  ]

  @classmethod
  def python_definition(cls, markdown: bool = True) -> str:
    """Returns the Python definition of the tool."""
    return lf_schema.Schema.from_value(cls).schema_str(
        protocol='python', markdown=markdown
    )

  def __call__(
      self,
      session,
      *,
      returns_call_result: bool = False) -> Any:
    """Calls a MCP tool synchronously.

    Args:
      session: A MCP session.
      returns_call_result: If True, returns the call result. Otherwise returns
        the result from structured content, or return content.

    Returns:
      The call result, or the result from structured content, or content.
    """
    return session.call_tool(
        self,
        returns_call_result=returns_call_result
    )

  async def acall(
      self,
      session,
      *,
      returns_call_result: bool = False) -> Any:
    """Calls a MCP tool asynchronously."""
    return await session.acall_tool(self, returns_call_result=returns_call_result)

  def input_parameters(self) -> dict[str, Any]:
    """Returns the input parameters of the tool."""
    json = self.to_json()
    def _transform(path: pg.KeyPath, x: Any) -> Any:
      del path
      if isinstance(x, dict):
        x.pop(pg.JSONConvertible.TYPE_NAME_KEY, None)
      return x
    return pg.utils.transform(json, _transform)

  @classmethod
  def make_class(cls, tool_definition: mcp.Tool) -> type['McpTool']:
    """Makes a MCP tool class from tool definition."""

    class _McpTool(cls):
      auto_schema = False

    tool_cls = _McpTool
    tool_cls.TOOL_NAME = tool_definition.name
    tool_cls.__name__ = _snake_to_camel(tool_definition.name)
    tool_cls.__doc__ = tool_definition.description
    schema = pg.Schema.from_json_schema(
        tool_definition.inputSchema, class_fn=McpToolInput.make_class
    )
    tool_cls.apply_schema(schema)
    return tool_cls


class _McpToolInputMeta(pg.symbolic.ObjectMeta):

  def __repr__(self) -> str:
    return f'<input-class \'{self.__name__}\'>'


class McpToolInput(pg.Object, metaclass=_McpToolInputMeta):
  """Base class for MCP tool inputs."""

  @classmethod
  def make_class(cls, name: str, schema: pg.Schema):
    """Converts a schema to an input class."""

    class _McpToolInput(cls):
      pass

    input_cls = _McpToolInput
    input_cls.__name__ = _snake_to_camel(name)
    input_cls.__doc__ = schema.description
    input_cls.apply_schema(schema)
    return input_cls


def _snake_to_camel(name: str) -> str:
  """Converts a snake_case name to a CamelCase name."""
  return ''.join(x.capitalize() for x in name.split('_'))



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import concurrent.futures
import time
import unittest

from langfun.env import interface
from langfun.env import load_balancers


class TestingSandbox(interface.Sandbox):
  sandbox_id: str
  status: interface.Sandbox.Status = interface.Sandbox.Status.READY

  def _on_bound(self) -> None:
    super()._on_bound()
    self._session_id = None

  @property
  def id(self) -> interface.Sandbox.Id:
    return interface.Sandbox.Id(
        environment_id=interface.Environment.Id('testing-env'),
        sandbox_id=self.sandbox_id
    )

  @property
  def environment(self) -> interface.Environment:
    raise NotImplementedError()

  @property
  def features(self) -> dict[str, interface.Feature]:
    raise NotImplementedError()

  @property
  def state_errors(self) -> list[interface.SandboxStateError]:
    return []

  def set_status(self, status: interface.Sandbox.Status) -> None:
    self.rebind(status=status, skip_notification=True)

  def set_acquired(self) -> None:
    self.set_status(self.Status.ACQUIRED)

  def start(self) -> None:
    self.set_alive()

  def shutdown(self) -> None:
    self.set_alive(False)

  def start_session(self, session_id: str) -> None:
    self._session_id = session_id

  def end_session(self, session_id: str) -> None:
    self._session_id = None

  @property
  def session_id(self) -> str | None:
    return self._session_id


class RoundRobinTest(unittest.TestCase):

  def test_basic(self):
    sandbox_pool = [
        TestingSandbox('0', interface.Sandbox.Status.OFFLINE),
        TestingSandbox('1', interface.Sandbox.Status.SETTING_UP),
        TestingSandbox('2', interface.Sandbox.Status.IN_SESSION),
        TestingSandbox('3', status=interface.Sandbox.Status.READY),
        TestingSandbox('4', status=interface.Sandbox.Status.READY),
    ]
    lb = load_balancers.RoundRobin()
    sandbox = lb.acquire(sandbox_pool)
    self.assertIs(sandbox, sandbox_pool[3])
    self.assertEqual(sandbox.status, interface.Sandbox.Status.ACQUIRED)

    sandbox = lb.acquire(sandbox_pool)
    self.assertIs(sandbox, sandbox_pool[4])
    self.assertEqual(sandbox.status, interface.Sandbox.Status.ACQUIRED)

    sandbox_pool[0].set_status(interface.Sandbox.Status.READY)
    sandbox = lb.acquire(sandbox_pool)
    self.assertIs(sandbox, sandbox_pool[0])
    self.assertEqual(sandbox.status, interface.Sandbox.Status.ACQUIRED)

    with self.assertRaisesRegex(IndexError, 'No free sandbox in the pool.'):
      lb.acquire(sandbox_pool)

  def test_thread_safety(self):
    sandbox_pool = [TestingSandbox(str(i)) for i in range(64)]

    lb = load_balancers.RoundRobin()

    def _thread_func(i):
      sandbox = lb.acquire(sandbox_pool)
      time.sleep(0.1)
      sandbox.set_status(interface.Sandbox.Status.IN_SESSION)
      time.sleep(0.1)
      sandbox.set_status(interface.Sandbox.Status.OFFLINE)
      time.sleep(0.1)
      sandbox.set_status(interface.Sandbox.Status.READY)
      return i

    with concurrent.futures.ThreadPoolExecutor(max_workers=64) as executor:
      for i, o in enumerate(executor.map(_thread_func, range(1024))):
        self.assertEqual(o, i)


if __name__ == '__main__':
  unittest.main()



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A simple MCP client for testing."""

from absl import app
from absl import flags
from langfun.core import mcp


_URL = flags.DEFINE_string(
    'url',
    'http://localhost:8000/mcp',
    'URL of the MCP server.',
)


def main(_):
  print(mcp.McpClient.from_url(url=_URL.value).list_tools())


if __name__ == '__main__':
  app.run(main)



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Environment event logger."""

import re
import time
from typing import Annotated
from langfun.env.event_handlers import base
import pyglove as pg


class EventLogger(pg.Object, base.EventHandler):
  """Event handler for logging debugger."""

  colored: Annotated[
      bool,
      (
          'If True, log events with colors.'
      )
  ] = False

  regex: Annotated[
      str | list[str] | None,
      (
          'One or a list of regular expressions to filter event messages. '
          'If None, no filtering will be applied.'
      )
  ] = None

  error_only: Annotated[
      bool,
      (
          'If True, log events with errors only.'
      )
  ] = False

  sandbox_status: Annotated[
      bool,
      (
          'If True, log events for sandbox status changes.'
      )
  ] = True

  feature_status: Annotated[
      bool,
      (
          'If True, log events for feature setup/teardown updates.'
      )
  ] = True

  session_status: Annotated[
      bool,
      (
          'If True, log events for session start/end status update.'
      )
  ] = True

  housekeep_status: Annotated[
      bool,
      (
          'If True, log housekeeping events.'
      )
  ] = True

  stats_report_interval: Annotated[
      float | None,
      (
          'The minimum interval in seconds for reporting the environment '
          'stats. If None, stats will not be reported.'
      )
  ] = 300.0

  def _on_bound(self) -> None:
    super()._on_bound()

    regex_exps = self.regex
    if isinstance(regex_exps, str):
      regex_exps = [regex_exps]
    elif regex_exps is None:
      regex_exps = []
    self._regex_exps = [re.compile(x) for x in regex_exps]
    self._last_stats_report_time = None

  def _maybe_colored(
      self, message: str, color: str, styles: list[str] | None = None
  ) -> str:
    if self.colored:
      return pg.colored(message, color, styles=styles)
    return message

  def _format_message(
      self,
      message: str,
      error: BaseException | None,
  ) -> str:
    if error is not None:
      message = (
          f'{message} with error: {pg.utils.ErrorInfo.from_exception(error)}'
      )
    return message

  def _keep(
      self,
      message: str,
      error: BaseException | None,
  ) -> bool:
    if error is None and self.error_only:
      return False
    if self._regex_exps and all(
        not exp.match(message) for exp in self._regex_exps
    ):
      return False
    return True

  def on_environment_starting(
      self,
      environment: base.Environment,
  ) -> None:
    """Called when the environment is starting."""
    self._print(
        f'[{environment.id}] environment starting',
        error=None,
        color='green',
        styles=['bold'],
    )

  def on_environment_shutting_down(
      self,
      environment: base.Environment,
      offline_duration: float,
  ) -> None:
    """Called when the environment is shutting down."""
    self._print(
        f'[{environment.id}] environment shutting down '
        f'(offline_duration={offline_duration:.2f} seconds)',
        error=None,
        color='green',
        styles=['bold'],
    )

  def on_environment_start(
      self,
      environment: base.Environment,
      duration: float,
      error: BaseException | None
  ) -> None:
    """Called when the environment is started."""
    self._print(
        f'[{environment.id}] environment started '
        f'(duration={duration:.2f} seconds)',
        error=error,
        color='green',
        styles=['bold'],
    )

  def on_environment_housekeep(
      self,
      environment: base.Environment,
      counter: int,
      duration: float,
      error: BaseException | None,
      **kwargs
  ) -> None:
    """Called when the environment is housekeeping."""
    if self.housekeep_status:
      self._print(
          f'[{environment.id}] environment housekeeping complete '
          f'(counter={counter}, duration={duration:.2f} seconds, '
          f'housekeep_info={kwargs})',
          error=error,
          color='green',
      )
    if (self.stats_report_interval is not None and
        (self._last_stats_report_time is None
         or time.time() - self._last_stats_report_time
         > self.stats_report_interval)):
      self._write_log(
          f'[{environment.id}] environment stats: {environment.stats()}',
          color='magenta',
          error=None,
      )
      self._last_stats_report_time = time.time()

  def on_environment_shutdown(
      self,
      environment: base.Environment,
      duration: float,
      lifetime: float,
      error: BaseException | None
  ) -> None:
    """Called when the environment is shutdown."""
    self._print(
        f'[{environment.id}] environment shutdown '
        f'(duration={duration:.2f} seconds), lifetime={lifetime:.2f} seconds)',
        error=error,
        color='green',
        styles=['bold'],
    )

  def on_sandbox_start(
      self,
      environment: base.Environment,
      sandbox: base.Sandbox,
      duration: float,
      error: BaseException | None
  ) -> None:
    if self.sandbox_status:
      self._print(
          f'[{sandbox.id}] sandbox started '
          f'(duration={duration:.2f} seconds)',
          error=error,
          color='white',
          styles=['bold'],
      )

  def on_sandbox_status_change(
      self,
      environment: base.Environment,
      sandbox: base.Sandbox,
      old_status: base.Sandbox.Status,
      new_status: base.Sandbox.Status,
      span: float
  ) -> None:
    if self.sandbox_status:
      self._print(
          f'[{sandbox.id}] {old_status.value} '
          f'({span:.2f} seconds) -> {new_status.value}',
          error=None,
          color='white',
      )

  def on_sandbox_shutdown(
      self,
      environment: base.Environment,
      sandbox: base.Sandbox,
      duration: float,
      lifetime: float,
      error: BaseException | None
  ) -> None:
    if self.sandbox_status:
      self._print(
          f'[{sandbox.id}] sandbox shutdown '
          f'(duration={duration:.2f} seconds), '
          f'lifetime={lifetime:.2f} seconds)',
          error=error,
          color='white',
          styles=['bold'],
      )

  def on_sandbox_housekeep(
      self,
      environment: base.Environment,
      sandbox: base.Sandbox,
      counter: int,
      duration: float,
      error: BaseException | None,
      **kwargs
  ) -> None:
    """Called when a sandbox feature is housekeeping."""
    if self.sandbox_status and self.housekeep_status:
      self._print(
          f'[{sandbox.id}] sandbox housekeeping complete '
          f'(counter={counter}, duration={duration:.2f} seconds, '
          f'housekeep_info={kwargs})',
          error=error,
          color='white',
      )

  def on_feature_setup(
      self,
      environment: base.Environment,
      sandbox: base.Sandbox,
      feature: base.Feature,
      duration: float,
      error: BaseException | None
  ) -> None:
    """Called when a sandbox feature is setup."""
    if self.feature_status:
      self._print(
          f'[{sandbox.id}/<idle>/{feature.name}] feature setup complete '
          f'(duration={duration:.2f} seconds)',
          error=error,
          color='white',
      )

  def on_feature_teardown(
      self,
      environment: base.Environment,
      sandbox: base.Sandbox,
      feature: base.Feature,
      duration: float,
      error: BaseException | None
  ) -> None:
    """Called when a sandbox feature is teardown."""
    if self.feature_status:
      self._print(
          f'[{sandbox.id}/<idle>/{feature.name}] feature teardown complete '
          f'(duration={duration:.2f} seconds)',
          error=error,
          color='white',
      )

  def on_feature_setup_session(
      self,
      environment: base.Environment,
      sandbox: base.Sandbox,
      feature: base.Feature,
      session_id: str | None,
      duration: float,
      error: BaseException | None
  ) -> None:
    """Called when a sandbox feature is setup."""
    if self.feature_status:
      self._print(
          f'[{sandbox.id}/{session_id or "<idle>"}/{feature.name}] '
          f'feature setup complete (duration={duration:.2f} seconds)',
          error=error,
          color='yellow',
      )

  def on_feature_teardown_session(
      self,
      environment: base.Environment,
      sandbox: base.Sandbox,
      feature: base.Feature,
      session_id: str,
      duration: float,
      error: BaseException | None
  ) -> None:
    """Called when a sandbox feature is teardown."""
    if self.feature_status:
      self._print(
          f'[{sandbox.id}/{session_id}>/{feature.name}] '
          f'feature teardown complete (duration={duration:.2f} seconds)',
          error=error,
          color='yellow',
      )

  def on_feature_housekeep(
      self,
      environment: base.Environment,
      sandbox: base.Sandbox,
      feature: base.Feature,
      counter: int,
      duration: float,
      error: BaseException | None,
      **kwargs
  ) -> None:
    """Called when a sandbox feature is housekeeping."""
    if self.feature_status and self.housekeep_status:
      self._print(
          f'[{sandbox.id}/<idle>/{feature.name}] feature housekeeping complete '
          f'(counter={counter}, (duration={duration:.2f} seconds, '
          f'housekeep_info={kwargs})',
          error=error,
          color='white',
      )

  def on_session_start(
      self,
      environment: base.Environment,
      sandbox: base.Sandbox,
      session_id: str,
      duration: float,
      error: BaseException | None
  ) -> None:
    """Called when a sandbox session starts."""
    if self.session_status:
      self._print(
          f'[{sandbox.id}/{session_id}] session started '
          f'(duration={duration:.2f} seconds)',
          error=error,
          color='blue',
      )

  def on_session_end(
      self,
      environment: base.Environment,
      sandbox: base.Sandbox,
      session_id: str,
      duration: float,
      lifetime: float,
      error: BaseException | None
  ) -> None:
    """Called when a sandbox session ends."""
    if self.session_status:
      self._print(
          f'[{sandbox.id}/{session_id}] session ended '
          f'(duration={duration:.2f} seconds), '
          f'lifetime={lifetime:.2f} seconds)',
          error=error,
          color='blue',
      )

  def on_sandbox_activity(
      self,
      name: str,
      environment: base.Environment,
      sandbox: base.Sandbox,
      feature: base.Feature | None,
      session_id: str | None,
      duration: float,
      error: BaseException | None,
      **kwargs
  ) -> None:
    """Called when a sandbox activity is performed."""
    del environment
    log_id = f'{sandbox.id}/{session_id or "<idle>"}'
    if feature is not None:
      log_id = f'{log_id}/{feature.name}'

    color = 'yellow' if session_id is None else 'cyan'
    self._print(
        f'[{log_id}] call {name!r} '
        f'(duration={duration:.2f} seconds, kwargs={kwargs}) ',
        error,
        color=color
    )

  def _print(
      self,
      message: str,
      error: BaseException | None,
      color: str | None = None,
      styles: list[str] | None = None,
  ):
    message = self._format_message(message, error)
    if not self._keep(message, error):
      return
    self._write_log(message, error, color, styles)

  def _write_log(
      self,
      message: str,
      error: BaseException | None,
      color: str | None = None,
      styles: list[str] | None = None,
  ):
    message = self._maybe_colored(
        message, color if error is None else 'red', styles=styles
    )
    if error is not None:
      pg.logging.error(message)
    else:
      pg.logging.info(message)


class ConsoleEventLogger(EventLogger):
  """Event handler for console debugger."""

  colored = True

  def _write_log(
      self,
      message: str,
      error: BaseException | None,
      color: str | None = None,
      styles: list[str] | None = None
  ):
    print(
        self._maybe_colored(
            message, color if error is None else 'red', styles=styles
        )
    )



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common base class for sandbox-based environments.

This module provides `BaseEnvironment`, a common base class for sandbox-based
environments that handles pooling, load balancing, and maintenance.

Note that:
- Environments do not have to inherit from this class, especially if features
  like pooling or load balancing are not needed.
- `BaseEnvironment` is coupled with `BaseSandbox`.
"""

import abc
import functools
import random
import threading
import time
from typing import Annotated, Any
import uuid

import langfun.core as lf
from langfun.env import base_sandbox
from langfun.env import interface
from langfun.env import load_balancers
from langfun.env.event_handlers import base as event_handler_base
import pyglove as pg


class BaseEnvironment(interface.Environment):
  """Common base for environments.

  The base environment provides the common functionalities for sandbox-based
  environments, such as environment pooling, load balancing, and sandbox
  maintenance.
  """

  root_dir: Annotated[
      str | None,
      (
          'The root directory for the environment for writting output files.'
          'If None, no output files will be allowed for the sandboxes.'
      )
  ] = None

  pool_size: Annotated[
      int | tuple[int, int],
      (
          'The (min_size, max_size) of the sandbox pool. If an integer, it '
          'will be used as both min and max size. If 0, sandboxes will be '
          'created on demand and shutdown when user session ends.'
      )
  ] = (0, 256)

  load_balancer: Annotated[
      load_balancers.LoadBalancer,
      (
          'The load balancer for the environment to acquire sandboxes.'
      )
  ] = load_balancers.RoundRobin()

  sandbox_keepalive_interval: Annotated[
      float | None,
      (
          'The interval in seconds to send keepalive pings to sandboxes. '
          'If None, sandbox keepalive is disabled. Please note that sandbox '
          'keepalive is different from feature housekeeping. Usually sandbox '
          'keepalive and feature housekeeping are different operations.'
      )
  ] = None

  proactive_session_setup: Annotated[
      bool,
      (
          'If True, all sandboxes will perform setup work before a user '
          'session is started. This is useful for features that need to '
          'perform heavy setup work, which could block the user thread for a '
          'long time.'
      )
  ] = True

  event_handlers: Annotated[
      list[event_handler_base.EventHandler],
      (
          'User handler for the environment events.'
      )
  ] = []

  outage_grace_period: Annotated[
      float,
      (
          'The grace period in seconds before the environment is treated '
          'as out of service. When calling `environment.sandbox()`, '
          'wait until the grace period has passed before raising an error.'
      )
  ] = 3600.0

  outage_retry_interval: Annotated[
      float,
      (
          'The retry interval in seconds for environment outage. '
          'When calling `environment.sandbox()`, retry after the interval '
          'if the environment is out of service.'
      )
  ] = 10.0

  housekeep_interval: Annotated[
      float,
      (
          'The interval in seconds for environment housekeeping. It recycles '
          'the dead sandboxes in the pool. This interval is the minimal time '
          'to detect outage while there is no request to obtain new sandboxes.'
          'This is applicable only when the environment enables pooling.'
      )
  ] = 10.0

  pool_operation_max_parallelism: Annotated[
      int,
      (
          'The maximum number of threads for bringing up or shutting down '
          'sandboxes in the pool.'
      )
  ] = 256

  random_seed: Annotated[
      int | None,
      (
          'The random seed for generating session IDs with reproducibility. '
          'If None, no seed will be used.'
      )
  ] = None

  def _on_bound(self) -> None:
    super()._on_bound()

    self._status = self.Status.CREATED
    self._start_time = None
    self._sandbox_pool = []
    self._next_pooled_sandbox_id = 0
    self._random = (
        random if self.random_seed is None else random.Random(self.random_seed)
    )

    self._housekeep_thread = None
    self._offline_start_time = None

  #
  # Subclasses must implement:
  #

  @abc.abstractmethod
  def _create_sandbox(
      self,
      sandbox_id: str,
      reusable: bool,
      proactive_session_setup: bool,
      keepalive_interval: float | None,
  ) -> base_sandbox.BaseSandbox:
    """Creates a sandbox with the given identifier.

    Args:
      sandbox_id: The identifier for the sandbox.
      reusable: Whether the sandbox is reusable across user sessions.
      proactive_session_setup: Whether the sandbox performs session setup work
        before a user session is started.
      keepalive_interval: Interval to ping the sandbox for keeping it alive.
        If None, the sandbox will not be pinged.

    Returns:
      The created sandbox.

    Raises:
      interface.EnvironmentError: If environment cannot create the sandbox.
      interface.SandboxStateError: If sandbox cannot be started.
    """

  def new_session_id(self) -> str:
    """Generates a random session ID."""
    suffix = uuid.UUID(
        bytes=bytes(bytes(self._random.getrandbits(8) for _ in range(16))),
        version=4
    ).hex[:7]
    return f'session-{suffix}'

  @property
  def housekeep_counter(self) -> int:
    """Returns the housekeeping counter."""
    return self._housekeep_counter

  #
  # Subclasses can override:
  #

  def stats(self) -> dict[str, Any]:
    """Returns the stats of the environment."""
    stats_dict = {
        status.value: 0
        for status in interface.Sandbox.Status
    }
    for sandbox in self._sandbox_pool:
      stats_dict[sandbox.status.value] += 1
    return {
        'sandbox': stats_dict,
    }

  def _start(self) -> None:
    """Implementation of starting the environment."""
    if self.min_pool_size > 0:
      # Pre-allocate the sandbox pool before usage.
      self._sandbox_pool = [None] * self.min_pool_size
      for i, sandbox, _ in lf.concurrent_map(
          lambda i: self._bring_up_sandbox_with_retry(
              sandbox_id=f'{i}:0', shutdown_env_upon_outage=False
          ),
          range(self.min_pool_size),
          silence_on_errors=None,
          max_workers=min(
              self.pool_operation_max_parallelism,
              self.min_pool_size
          ),
      ):
        self._sandbox_pool[i] = sandbox

    self._next_sandbox_id = len(self._sandbox_pool)

    if self.enable_pooling:
      self._housekeep_thread = threading.Thread(
          target=self._housekeep_loop, daemon=True
      )
      self._housekeep_counter = 0
      self._housekeep_thread.start()

  def _shutdown(self) -> None:
    """Implementation of shutting down the environment."""
    if (self._housekeep_thread is not None
        and threading.current_thread() is not self._housekeep_thread):
      self._housekeep_thread.join()
      self._housekeep_thread = None

    def _shutdown_sandbox(sandbox: base_sandbox.BaseSandbox) -> None:
      if sandbox is not None:
        sandbox.shutdown()

    if self._sandbox_pool:
      _ = list(
          lf.concurrent_map(
              _shutdown_sandbox,
              self._sandbox_pool,
              silence_on_errors=None,
              max_workers=min(
                  self.pool_operation_max_parallelism,
                  len(self._sandbox_pool)
              ),
          )
      )
      self._sandbox_pool = []

  #
  # Environment basics.
  #

  @property
  def sandbox_pool(self) -> list[base_sandbox.BaseSandbox]:
    """Returns the sandbox pool."""
    return self._sandbox_pool

  @functools.cached_property
  def working_dir(self) -> str | None:
    """Returns the working directory for the environment."""
    return self.id.working_dir(self.root_dir)

  @property
  def enable_pooling(self) -> bool:
    """Returns whether the environment enables pooling."""
    return self.max_pool_size > 0

  @property
  def status(self) -> interface.Environment.Status:
    """Returns whether the environment is online."""
    return self._status

  def _set_status(self, status: interface.Environment.Status) -> None:
    """Sets the status of the environment."""
    self._status = status

  @property
  def min_pool_size(self) -> int:
    """Returns the minimum size of the sandbox pool."""
    if isinstance(self.pool_size, int):
      return self.pool_size
    return self.pool_size[0]

  @property
  def max_pool_size(self) -> int:
    """Returns the maximum size of the sandbox pool."""
    if isinstance(self.pool_size, int):
      return self.pool_size
    return self.pool_size[1]

  @property
  def start_time(self) -> float | None:
    """Returns the start time of the environment."""
    return self._start_time

  @property
  def offline_duration(self) -> float:
    """Returns the offline duration of the environment."""
    if self._offline_start_time is None:
      return 0.0
    return time.time() - self._offline_start_time

  #
  # Environment lifecycle.
  #

  def start(self) -> None:
    """Starts the environment.

    Raises:
      interface.EnvironmentOutageError: If the environment is out of service.
    """
    assert self._status == self.Status.CREATED, (
        f'Environment {self.id} cannot be started because '
        f'it is in {self._status.value!r} status.'
    )

    self.on_starting()
    starting_time = time.time()
    try:
      self._start()
      self._start_time = time.time()
      self._set_status(self.Status.ONLINE)
      self.on_start(duration=time.time() - starting_time)
    except BaseException as e:
      self.on_start(duration=time.time() - starting_time, error=e)
      self.shutdown()
      raise e

  def shutdown(self) -> None:
    """Shuts down the environment.

    This method should not raise any exceptions.
    """
    if self._status in (
        self.Status.SHUTTING_DOWN,
        self.Status.OFFLINE,
    ):
      return

    self._set_status(self.Status.SHUTTING_DOWN)
    self.on_shutting_down()

    shutting_down_time = time.time()
    try:
      self._shutdown()
      self.on_shutdown(duration=time.time() - shutting_down_time)
    except BaseException as e:  # pylint: disable=broad-except
      self.on_shutdown(duration=time.time() - shutting_down_time, error=e)
      raise e

  #
  # Environment operations.
  #

  def acquire(self) -> base_sandbox.BaseSandbox:
    """Acquires a sandbox from the environment.

    Returns:
      The acquired sandbox.

    Raises:
      interface.EnvironmentOutageError: If the environment is offline and the
        grace period has passed.
      interface.EnvironmentOverloadError: If the max pool size is reached and
        the grace period has passed.
    """

    if not self.is_online:
      raise interface.EnvironmentOutageError(
          f'Environment {self.id} is not alive.',
          environment=self,
          offline_duration=self.offline_duration,
      )

    if not self.enable_pooling:
      return self._bring_up_sandbox_with_retry(
          sandbox_id=str(self._increment_sandbox_id()),
          set_acquired=True,
      )

    allocation_start_time = time.time()
    while True:
      try:
        # We only append or replace items in the sandbox pool, therefore
        # there is no need to lock the pool.
        return self.load_balancer.acquire(self._sandbox_pool)
      except IndexError:
        if len(self._sandbox_pool) == self.max_pool_size:
          if time.time() - allocation_start_time > self.outage_grace_period:
            raise interface.EnvironmentOverloadError(  # pylint: disable=raise-missing-from
                environment=self
            )
          time.sleep(1)
        else:
          try:
            sandbox = self._bring_up_sandbox(
                sandbox_id=f'{self._increment_sandbox_id()}:0',
                set_acquired=True,
            )
            # Append is atomic and does not require locking.
            self._sandbox_pool.append(sandbox)
            return sandbox
          except (
              interface.EnvironmentError, interface.SandboxStateError
          ) as ex:
            self._report_outage_or_wait(ex)

  def _bring_up_sandbox(
      self,
      sandbox_id: str,
      set_acquired: bool = False,
  ) -> base_sandbox.BaseSandbox:
    """Brings up a new sandbox."""
    env_error = None
    try:
      sandbox = self._create_sandbox(
          sandbox_id=sandbox_id,
          reusable=self.enable_pooling,
          proactive_session_setup=self.proactive_session_setup,
          keepalive_interval=self.sandbox_keepalive_interval,
      )
      for handler in self.event_handlers:
        sandbox.add_event_handler(handler)
      sandbox.start()
      if set_acquired:
        sandbox.set_acquired()
      return sandbox
    except (interface.EnvironmentError, interface.SandboxStateError) as e:
      env_error = e
      raise e
    finally:
      if env_error is None:
        self._offline_start_time = None
      elif self._offline_start_time is None:
        self._offline_start_time = time.time()

  def _bring_up_sandbox_with_retry(
      self,
      sandbox_id: str,
      set_acquired: bool = False,
      shutdown_env_upon_outage: bool = True,
  ) -> base_sandbox.BaseSandbox:
    """Brings up a new sandbox with retry until grace period is passed.

    Args:
      sandbox_id: The ID of the sandbox to bring up.
      set_acquired: If True, the sandbox will be marked as acquired.
      shutdown_env_upon_outage: Whether to shutdown the environment when the
        outage grace period is passed.

    Returns:
      A new sandbox ready to use.

    Raises:
      interface.EnvironmentOutageError: If the environment is offline and the
        grace period has passed.
    """
    while True:
      try:
        return self._bring_up_sandbox(
            sandbox_id=sandbox_id, set_acquired=set_acquired
        )
      except (interface.EnvironmentError, interface.SandboxStateError) as e:
        self._report_outage_or_wait(e, shutdown_env_upon_outage)

  def _increment_sandbox_id(self) -> int:
    """Returns the next pooled sandbox ID."""
    x = self._next_sandbox_id
    self._next_sandbox_id += 1
    return x

  def _report_outage_or_wait(
      self,
      error: interface.SandboxStateError,
      shutdown_env_upon_outage: bool = True
  ):
    """Raises error if the grace period has passed or wait for retry."""
    if self.offline_duration > self.outage_grace_period:
      if shutdown_env_upon_outage:
        self.shutdown()
      raise interface.EnvironmentOutageError(
          environment=self,
          offline_duration=self.offline_duration,
      ) from error
    time.sleep(self.outage_retry_interval)

  #
  # Environment maintenance loop.
  #

  def _housekeep_loop(self) -> None:
    """Housekeeping loop for the environment."""
    while self._status not in (self.Status.SHUTTING_DOWN, self.Status.OFFLINE):
      housekeep_start_time = time.time()

      is_online = True
      dead_pool_indices = [
          i for i, s in enumerate(self._sandbox_pool)
          if s.status == interface.Sandbox.Status.OFFLINE
      ]
      replaced_indices = []

      if dead_pool_indices:
        replaced_indices = self._replace_dead_sandboxes(dead_pool_indices)
        if not replaced_indices:
          is_online = self.offline_duration < self.outage_grace_period

      self._housekeep_counter += 1
      duration = time.time() - housekeep_start_time
      kwargs = dict(
          dead_pool_indices=dead_pool_indices,
          replaced_indices=replaced_indices,
          offline_duration=self.offline_duration,
      )
      if is_online:
        self.on_housekeep(duration, **kwargs)
        time.sleep(self.housekeep_interval)
      else:
        self.shutdown()
        self.on_housekeep(
            duration,
            interface.EnvironmentOutageError(
                environment=self, offline_duration=self.offline_duration
            ),
            **kwargs
        )

  def _replace_dead_sandboxes(self, dead_pool_indices: list[int]) -> list[int]:
    """Replaces a dead sandbox with a new one.

    Args:
      dead_pool_indices: The indices of the dead sandboxes to replace.

    Returns:
      Successfully replaced indices.
    """
    pg.logging.warning(
        '[%s]: %s maintenance: '
        'Replacing %d dead sandbox(es) with new ones...',
        self.id,
        self.__class__.__name__,
        len(dead_pool_indices),
    )
    def _replace(i: int):
      generation = int(self._sandbox_pool[i].id.sandbox_id.split(':')[1])
      self._sandbox_pool[i] = self._bring_up_sandbox(f'{i}:{generation + 1}')

    # TODO(daiyip): Consider to loose the condition to allow some dead
    # sandboxes to be replaced successfully.
    replaced_indices = []
    for index, _, error in lf.concurrent_map(
        _replace, dead_pool_indices,
        max_workers=min(
            self.pool_operation_max_parallelism,
            len(dead_pool_indices)
        ),
    ):
      if error is None:
        replaced_indices.append(index)

    pg.logging.warning(
        '[%s]: %s maintenance: '
        '%d/%d dead sandbox(es) have been replaced with new ones. (slots=%s)',
        self.id,
        self.__class__.__name__,
        len(replaced_indices),
        len(dead_pool_indices),
        replaced_indices
    )
    return replaced_indices

  #
  # Event handlers subclasses can override.
  #

  def on_starting(self) -> None:
    """Called when the environment is getting started."""
    for handler in self.event_handlers:
      handler.on_environment_starting(self)

  def on_start(
      self,
      duration: float, error: BaseException | None = None
  ) -> None:
    """Called when the environment is started."""
    for handler in self.event_handlers:
      handler.on_environment_start(self, duration, error)

  def on_housekeep(
      self,
      duration: float,
      error: BaseException | None = None,
      **kwargs
  ) -> None:
    """Called when the environment finishes a round of housekeeping."""
    housekeep_counter = self.housekeep_counter
    for handler in self.event_handlers:
      handler.on_environment_housekeep(
          self, housekeep_counter, duration, error, **kwargs
      )

  def on_shutting_down(self) -> None:
    """Called when the environment is shutting down."""
    for handler in self.event_handlers:
      handler.on_environment_shutting_down(self, self.offline_duration)

  def on_shutdown(
      self,
      duration: float,
      error: BaseException | None = None) -> None:
    """Called when the environment is shutdown."""
    lifetime = (time.time() - self.start_time) if self.start_time else 0.0
    for handler in self.event_handlers:
      handler.on_environment_shutdown(self, duration, lifetime, error)



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Simple MCP server for testing."""

from absl import app as absl_app
from mcp.server import fastmcp as fastmcp_lib

mcp = fastmcp_lib.FastMCP(host='0.0.0.0', port=8000)


@mcp.tool()
async def add(a: int, b: int) -> int:
  """Adds two integers and returns their sum."""
  return a + b


def main(_):
  mcp.run(transport='streamable-http', mount_path='/mcp')


if __name__ == '__main__':
  absl_app.run(main)



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interfaces for environments, sandboxes and features."""

import abc
import contextlib
import dataclasses
import enum
import os
from typing import Annotated, Any, ContextManager, ClassVar, Iterator, Optional

import pyglove as pg

#
# Environment errors.
#


class EnvironmentError(RuntimeError):  # pylint: disable=redefined-builtin
  """Base class for environment errors."""

  def __init__(
      self,
      message: str,
      *args,
      environment: 'Environment',
      **kwargs
  ) -> None:
    self.environment = environment
    super().__init__(f'[{environment.id}] {message}.', *args, **kwargs)


class EnvironmentOutageError(EnvironmentError):
  """Error that indicates environment is offline."""

  def __init__(
      self,
      message: str | None = None,
      *args,
      offline_duration: float,
      **kwargs
  ):
    self.offline_duration = offline_duration
    super().__init__(
        message or f'Environment is offline for {offline_duration} seconds.',
        *args,
        **kwargs
    )


class EnvironmentOverloadError(EnvironmentError):
  """Error that indicates environment is overloaded."""

  def __init__(
      self,
      message: str | None = None,
      *args,
      **kwargs
  ):
    super().__init__(
        message or 'All sandboxes in the pool are either busy or dead.',
        *args, **kwargs
    )


class SandboxError(RuntimeError):
  """Base class for sandbox errors."""

  def __init__(
      self,
      message: str,
      *args,
      sandbox: 'Sandbox',
      **kwargs
  ) -> None:
    self.sandbox = sandbox
    super().__init__(f'[{sandbox.id}] {message}.', *args, **kwargs)


class SandboxStateError(SandboxError):
  """Error that indicates sandbox is in an unexpected state.

  This error is raised when the sandbox is in an unexpected state and cannot
  be recovered. As a result, the sandbox will be shutdown and user session
  will be terminated.
  """

  def __init__(
      self,
      message: str | None = None,
      *args,
      code: str | None = None,
      **kwargs
  ):
    default_message = 'Sandbox is in an unexpected state'
    if code is not None:
      default_message = (
          f'Sandbox is in an unexpected state after executing code: {code!r}'
      )
    super().__init__(message or default_message, *args, **kwargs)


class FeatureTeardownError(SandboxError):
  """Base class for feature errors."""

  def __init__(
      self,
      message: str | None = None,
      *args,
      errors: dict[str, BaseException],
      **kwargs
  ):
    self.errors = errors
    super().__init__(
        (message or
         f'Feature teardown failed with user-defined errors: {errors}.'),
        *args,
        **kwargs
    )

  @property
  def has_non_sandbox_state_error(self) -> bool:
    """Returns True if the feature teardown error has non-sandbox state error."""
    return any(
        not isinstance(e, SandboxStateError) for e in self.errors.values()
    )


class SessionTeardownError(SandboxError):
  """Base class for session errors."""

  def __init__(
      self,
      message: str | None = None,
      *args,
      errors: dict[str, BaseException],
      **kwargs
  ):
    self.errors = errors
    super().__init__(
        (message or
         f'Session teardown failed with user-defined errors: {errors}.'),
        *args,
        **kwargs
    )

  @property
  def has_non_sandbox_state_error(self) -> bool:
    """Returns True if the feature teardown error has non-sandbox state error."""
    return any(
        not isinstance(e, SandboxStateError) for e in self.errors.values()
    )

#
# Interface for sandbox-based environment.
#


class Environment(pg.Object):
  """Base class for an environment."""

  @dataclasses.dataclass(frozen=True)
  class Id:
    """Identifier for an environment."""
    environment_id: str

    def __str__(self) -> str:
      return self.environment_id

    def working_dir(self, root_dir: str | None) -> str | None:
      """Returns the download directory for the service."""
      if root_dir is None:
        return None
      return os.path.join(root_dir, _make_path_compatible(self.environment_id))

  class Status(enum.Enum):
    """Environment state.

    State transitions:

    +---------------+               +-----------+
    |  <CREATED>    | --(start)---> |  ONLINE   | -(shutdown or outage detected)
    +---------------+      |        +-----------+              |
                           |        +-----------+              |
                           +------> | <OFFLINE> | <------------+
                                    +-----------+
    """
    CREATED = 'created'
    ONLINE = 'online'
    SHUTTING_DOWN = 'shutting_down'
    OFFLINE = 'offline'

  features: Annotated[
      dict[str, 'Feature'],
      'Features to be exposed by the environment.'
  ] = {}

  _ENV_STACK: Annotated[
      ClassVar[list['Environment']],
      'Recording the environments stacked through context managers.'
  ] = []

  #
  # Subclasses must implement:
  #

  @property
  @abc.abstractmethod
  def id(self) -> Id:
    """Returns the identifier for the environment."""

  @property
  @abc.abstractmethod
  def status(self) -> Status:
    """Returns the status of the environment."""

  @abc.abstractmethod
  def stats(self) -> dict[str, Any]:
    """Returns the stats of the environment."""

  @abc.abstractmethod
  def start(self) -> None:
    """Starts the environment.

    Raises:
      EnvironmentError: If the environment is not available.
    """

  @abc.abstractmethod
  def shutdown(self) -> None:
    """Shuts down the environment.

    IMPORTANT: This method shall not raise any exceptions.
    """

  @abc.abstractmethod
  def acquire(self) -> 'Sandbox':
    """Acquires a free sandbox from the environment.

    Returns:
      A free sandbox from the environment.

    Raises:
      EnvironmentOutageError: If the environment is out of service.
      EnvironmentOverloadError: If the environment is overloaded.
    """

  @abc.abstractmethod
  def new_session_id(self) -> str:
    """Generates a new session ID."""

  #
  # Environment lifecycle.
  #

  @property
  def is_online(self) -> bool:
    """Returns True if the environment is alive."""
    return self.status == self.Status.ONLINE

  def __enter__(self) -> 'Environment':
    """Enters the environment and sets it as the current environment."""
    self.start()
    Environment._ENV_STACK.append(self)
    return self

  def __exit__(self, exc_type, exc_value, traceback):
    """Exits the environment and reset the current environment."""
    assert Environment._ENV_STACK
    Environment._ENV_STACK.pop()
    self.shutdown()

  def __del__(self):
    """Deletes the environment."""
    self.shutdown()

  @classmethod
  def current(cls) -> Optional['Environment']:
    """Returns the current environment."""
    if not Environment._ENV_STACK:
      return None
    return Environment._ENV_STACK[-1]

  #
  # Environment operations.
  #

  def sandbox(
      self,
      session_id: str | None = None,
  ) -> ContextManager['Sandbox']:
    """Gets a sandbox from the environment and starts a new user session."""
    return self.acquire().new_session(session_id)

  def __getattr__(self, name: str) -> Any:
    """Gets a feature from a free sandbox from the environment.

    Example:
      ```
      with XboxEnvironment(
          features={'selenium': SeleniumFeature()}
      ) as env:
        driver = env.selenium.get_driver()
      ```

    Args:
      name: The name of the feature.

    Returns:
      A feature from a free sandbox from the environment.
    """
    if name in self.features:
      return self.acquire().features[name]
    raise AttributeError(name)


# Enable automatic conversion from str to Environment.Id.
pg.typing.register_converter(str, Environment.Id, Environment.Id)


class Sandbox(pg.Object):
  """Interface for sandboxes."""

  @dataclasses.dataclass(frozen=True, slots=True)
  class Id:
    """Identifier for a sandbox."""
    environment_id: Environment.Id
    sandbox_id: str

    def __str__(self) -> str:
      return f'{self.environment_id}/{self.sandbox_id}'

    def working_dir(self, root_dir: str | None) -> str | None:
      """Returns the download directory for the sandbox."""
      if root_dir is None:
        return None
      return os.path.join(
          self.environment_id.working_dir(root_dir),
          _make_path_compatible(self.sandbox_id)
      )

  class Status(enum.Enum):
    r"""Sandbox state.

    State transitions:

                    (sandbox / feature
    +------------+   teardown)            +---------------+
    | <OFFLINE>  | <--------------------- | SHUTTING_DOWN |
    +------------+                        +---------------+
                                            ^     ^
                                           /       \
                          (setup failed)  /         \
                                         /           \
    +-----------+   (start)  +------------+           \
    | <CREATED> | -------->  | SETTING_UP |            \
    +-----------+         ^  +------------+             \
                         /        |                      \
                        /         | (sandbox /            \
                       /          |  feature /session      \
                      /           v  setup succeeded)       \
                     /        +---------+                    \
                    /         |  READY  |                     \
                   /          +---------+                      \
                  /                |                            \
                 /                 |  (acquire)                  \
                /                  v                              \
               /              +----------+                         \
              |               | ACQUIRED |                          \
              |               +----------+                           |
              |                    |                                 |
              |                    |  (start_session)                |
              |               +------------+                         |
              |               | SETTING_UP |-- (setup failed) ------>+
              |               +------------+                         |
              |                    |                                 |
              |                    v  (succeeded)                    |
              |               +--------------+                       |
              |               |  IN_SESSION  |- (op failed) -------->+
              |               +--------------+                       |
              |                    |                                 |
              |                    |  (end_session)                  |
              |                    |                                 |
              |                    v             (session teardown   |
      (setup next            +-----------------+  failed OR          |
      session for  <---------| EXITING_SESSION |- non-reusable  -----+
      reusable sandbox)      +-----------------+  sandbox)

    """

    # The sandbox is created, but not yet started.
    CREATED = 'created'

    # The sandbox is being setting up to serve user sessions.
    SETTING_UP = 'setting_up'

    # The sandbox is set up and free to be acquired by the user.
    READY = 'ready'

    # The sandbox is acquired by a thread, but not yet in a user session.
    ACQUIRED = 'acquired'

    # The sandbox is in a user session.
    IN_SESSION = 'in_session'

    # The sandbox is exiting a user session.
    EXITING_SESSION = 'exiting_session'

    # The sandbox is being shut down.
    SHUTTING_DOWN = 'shutting_down'

    # The sandbox is offline.
    OFFLINE = 'offline'

    @property
    def is_online(self) -> bool:
      """Returns True if the sandbox is online."""
      return self in (
          Sandbox.Status.SETTING_UP,
          Sandbox.Status.READY,
          Sandbox.Status.ACQUIRED,
          Sandbox.Status.IN_SESSION,
      )

  @property
  @abc.abstractmethod
  def id(self) -> Id:
    """Returns the identifier for the sandbox."""

  @property
  @abc.abstractmethod
  def environment(self) -> Environment:
    """Returns the environment for the sandbox."""

  @property
  @abc.abstractmethod
  def features(self) -> dict[str, 'Feature']:
    """Returns the features in the sandbox."""

  @property
  @abc.abstractmethod
  def status(self) -> Status:
    """Returns the status of the sandbox."""

  @property
  def is_online(self) -> bool:
    """Returns True if the sandbox is online."""
    return self.status.is_online

  @property
  @abc.abstractmethod
  def state_errors(self) -> list[SandboxStateError]:
    """Returns state errors encountered during sandbox's lifecycle."""

  @abc.abstractmethod
  def start(self) -> None:
    """Starts the sandbox.

    State transitions:
      CREATED -> SETTING_UP -> READY: When all sandbox and feature setup
        succeeds.
      CREATED -> SETTING_UP -> SHUTTING_DOWN -> OFFLINE: When sandbox or feature
        setup fails.

    `start` and `shutdown` should be called in pairs, even when the sandbox
    fails to start. This ensures proper cleanup.

    Start may fail with two sources of errors:

    1. SandboxStateError: If sandbox or feature setup fail due to enviroment
       outage or sandbox state errors.
    2. BaseException: If feature setup failed with user-defined errors, this
       could happen when there is bug in the user code or non-environment code
       failure.

    In both cases, the sandbox will be shutdown automatically, and the error
    will be add to `errors`. The sandbox is considered dead and will not be
    further used.

    Raises:
      SandboxStateError: If the sandbox is in a bad state.
      BaseException: If feature setup failed with user-defined errors.
    """

  @abc.abstractmethod
  def shutdown(self) -> None:
    """Shuts down the sandbox.

    State transitions:
      SHUTTING_DOWN -> SHUTTING_DOWN: No operation.
      OFFLINE -> OFFLINE: No operation.
      SETTING_UP -> SHUTTING_DOWN -> OFFLINE: When sandbox and feature
        setup fails.
      IN_SESSION -> SHUTTING_DOWN -> OFFLINE: When user session exits while
        sandbox is set not to reuse, or session teardown fails.
      FREE -> SHUTTING_DOWN -> OFFLINE: When sandbox is shutdown when the
        environment is shutting down, or housekeeping loop shuts down the
        sandbox due to housekeeping failures.


    Please be aware that `shutdown` will be called whenever an operation on the
    sandbox encounters a critical error. This means, `shutdown` should not make
    the assumption that the sandbox is in a healthy state, even `start` could
    fail. As a result, `shutdown` must allow re-entry and be thread-safe with
    other sandbox operations.

    Shutdown may fail with two sources of errors:

    1. SandboxStateError: If the sandbox is in a bad state, and feature teardown
       logic depending on a healthy sandbox may fail. In such case, we do not
       raise error to the user as the user session is considered completed. The
       sandbox is abandoned and new user sessions will be served on other
       sandboxes.

    2. BaseException: The sandbox is in good state, but user code raises error
       due to bug or non-environment code failure. In such case, errors will be
       raised to the user so the error could be surfaced and handled properly.
       The sandbox is treated as shutdown and will not be further used.

    Raises:
      BaseException: If feature teardown failed with user-defined errors.
    """

  @abc.abstractmethod
  def start_session(
      self,
      session_id: str,
  ) -> None:
    """Begins a user session with the sandbox.

    State transitions:
      ACQUIRED -> SETTING_UP -> IN_SESSION: When session setup succeeds.
      ACQUIRED -> SETTING_UP -> SHUTTING_DOWN -> OFFLINE: When session setup
        fails.

    A session is a sequence of stateful interactions with the sandbox.
    Across different sessions the sandbox are considered stateless.
    `start_session` and `end_session` should always be called in pairs, even
    when the session fails to start. `Sandbox.new_session` context manager is
    the recommended way to use `start_session` and `end_session` in pairs.

    Starting a session may fail with two sources of errors:

    1. SandboxStateError: If the sandbox is in a bad state or session setup
       failed.

    2. BaseException: If session setup failed with user-defined errors.

    In both cases, the sandbox will be shutdown automatically and the
    session will be considered ended. The error will be added to `errors`.
    Future session will be served on other sandboxes.

    Args:
      session_id: The identifier for the user session.

    Raises:
      SandboxStateError: If the sandbox is already in a bad state or session
        setup failed.
      BaseException: If session setup failed with user-defined errors.
    """

  @abc.abstractmethod
  def end_session(self) -> None:
    """Ends the user session with the sandbox.

    State transitions:
      IN_SESSION -> EXITING_SESSION -> READY: When user session exits normally,
        and sandbox is set to reuse.
      IN_SESSION -> EXITING_SESSION -> SHUTTING_DOWN -> OFFLINE: When user
        session exits while
        sandbox is set not to reuse, or session teardown fails.
      IN_SESSION -> EXITING_SESSION -> SETTING_UP -> READY: When user session
        exits normally, and sandbox is set to reuse, and proactive session setup
        is enabled.
      IN_SESSION -> EXITING_SESSION -> SETTING_UP -> SHUTTING_DOWN -> OFFLINE:
        When user session exits normally, and proactive session setup is enabled
        but fails.
      EXITING_SESSION -> EXITING_SESSION: No operation.
      not IN_SESSION -> same state: No operation

    `end_session` should always be called for each `start_session` call, even
    when the session fails to start, to ensure proper cleanup.

    `end_session` may fail with two sources of errors:

    1. SandboxStateError: If the sandbox is in a bad state or session teardown
       failed.

    2. BaseException: If session teardown failed with user-defined errors.

    In both cases, the sandbox will be shutdown automatically and the
    session will be considered ended. The error will be added to `errors`.
    Future session will be served on other sandboxes.

    However, SandboxStateError encountered during `end_session` will NOT be
    raised to the user as the user session is considered completed.

    Raises:
      BaseException: If session teardown failed with user-defined errors.
    """

  @property
  @abc.abstractmethod
  def session_id(self) -> str | None:
    """Returns the current user session identifier."""

  #
  # API related to a user session.
  # A sandbox could be reused across different user sessions.
  # A user session is a sequence of stateful interactions with the sandbox,
  # Across different sessions the sandbox are considered stateless.
  #

  @contextlib.contextmanager
  def new_session(
      self,
      session_id: str | None = None,
  ) -> Iterator['Sandbox']:
    """Context manager for obtaining a sandbox for a user session.

    State transitions:
      ACQUIRED -> IN_SESSION -> READY: When session setup and teardown succeed.
      ACQUIRED -> IN_SESSINO -> OFFLINE: When session setup or teardown fails.

    Args:
      session_id: The identifier for the user session. If not provided, a random
        ID will be generated.

    Yields:
      The sandbox for the user session.

    Raises:
      SandboxStateError: If a session cannot be started on the sandbox.
      BaseException: If session setup or teardown failed with user-defined
        errors.
    """
    if session_id is None:
      session_id = self.environment.new_session_id()
    self.start_session(session_id)
    try:
      yield self
    finally:
      self.end_session()

  def __getattr__(self, name: str) -> Any:
    """Gets a feature from current sandbox.

    Example:
      ```
      with MyEnvironment(
          features={'feature1': Feature1()}
      ) as env:
        with env.sandbox('session1') as sb:
          driver = sb.feature1.feature_method()
      ```

    Args:
      name: The name of the feature.

    Returns:
      A feature from current sandbox.
    """
    if name in self.features:
      return self.features[name]
    raise AttributeError(name)

  def __del__(self):
    """Deletes the sandbox."""
    self.shutdown()


class Feature(pg.Object):
  """Interface for sandbox features."""

  @property
  @abc.abstractmethod
  def name(self) -> str:
    """Name of the feature, which will be used as key to access the feature."""

  @property
  @abc.abstractmethod
  def sandbox(self) -> Sandbox:
    """Returns the sandbox that the feature is running in.

    Returns:
      The sandbox that the feature is running in.

    Raises:
      AssertError: If the feature is not set up with a sandbox yet.
    """

  @abc.abstractmethod
  def setup(self, sandbox: Sandbox) -> None:
    """Sets up the feature, which is called once when the sandbox is up.

    State transitions:
      SETTING_UP -> READY: When setup succeeds.
      SETTING_UP -> SHUTTING_DOWN -> OFFLINE: When setup fails.

    `setup` is called when a sandbox is started for the first time. When a
    feature's `setup` is called, its `teardown` is guaranteed to be called.

    Args:
      sandbox: The sandbox that the feature is running in.

    Raises:
      SandboxStateError: If setup failed due to sandbox state errors.
      BaseException: If setup failed with user-defined errors.
    """

  @abc.abstractmethod
  def teardown(self) -> None:
    """Teardowns the feature, which is called once when the sandbox is down.

    State transitions:
      SHUTTING_DOWN -> OFFLINE: When teardown succeeds or fails.

    Raises:
      SandboxStateError: If the sandbox is in a bad state.
    """

  @abc.abstractmethod
  def setup_session(self) -> None:
    """Sets up the feature for the upcoming user session.

    State transitions:
      SETTING_UP -> READY: When session setup succeeds.
      SETTING_UP -> SHUTTING_DOWN -> OFFLINE: When sessino setup fails.

    `setup_session` is called when a new user session starts for per-session
    setup. `setup_session` and `teardown_session` will be called in pairs, even
    `setup_session` fails.

    `setup_session` may fail with two sources of errors:

    1. SandboxStateError: If the sandbox is in a bad state or session setup
       failed.
    2. BaseException: If session setup failed with user-defined errors.

    In both cases, the error will be raised to the user and the session will be
    ended. The sandbox will shutdown automatically and will not be further used.

    Raises:
      SandboxStateError: If the sandbox is in a bad state or session setup
        failed.
      BaseException: If session setup failed with user-defined errors.
    """

  @abc.abstractmethod
  def teardown_session(self) -> None:
    """Teardowns the feature for an ending user session.

    State transitions:
      SHUTTING_DOWN -> OFFLINE: When session teardown succeeds or fails.

    `teardown_session` is called when a user session ends for per-
    session teardown. `teardown_session` will always be called upon a feature
    whose `setup_session` is called.

    `teardown_session` may fail with two sources of errors:

    1. SandboxStateError: If the sandbox is in a bad state or session teardown
       failed.
    2. BaseException: If session teardown failed with user-defined errors.

    In both cases, the session will be closed and the sandbox will be shutdown.
    However, SandboxStateError encountered during `teardown_session` will NOT be
    raised to the user as the user session is considered completed. Other errors
    will be raised to the user for proper error handling.

    Raises:
      BaseException: If session teardown failed with user-defined errors.
    """

  @abc.abstractmethod
  def housekeep(self) -> None:
    """Performs housekeeping for the feature.

    State transitions:
      (no state change): When housekeeping succeeds.
      original state -> SHUTTING_DOWN -> OFFLINE: When housekeeping fails.

    Raises:
      SandboxStateError: If the sandbox is in a bad state.
    """

  @property
  @abc.abstractmethod
  def housekeep_interval(self) -> int | None:
    """Returns the interval in seconds for feature housekeeping.

    Returns:
      The interval in seconds for feature housekeeping. If None, the feature
      will not be housekeeping.
    """

  @property
  def session_id(self) -> str | None:
    """Returns the current user session identifier."""
    assert self.sandbox is not None
    return self.sandbox.session_id


def _make_path_compatible(id_str: str) -> str:
  """Makes a path compatible with CNS."""
  return id_str.translate(
      str.maketrans({
          '@': '_',
          ':': '_',
          '#': '_',
          ' ': '',
          '<': '',
          '>': '',
      })
  )



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for MCP client."""

import inspect
import unittest
from langfun.core import async_support
from langfun.core import mcp as lf_mcp
from mcp.server import fastmcp as fastmcp_lib

mcp = fastmcp_lib.FastMCP(host='0.0.0.0', port=1234)


@mcp.tool()
async def add(a: int, b: int) -> int:
  """Adds two integers and returns their sum.

  Args:
    a: The first integer.
    b: The second integer.

  Returns:
    The sum of the two integers.
  """
  return a + b


class McpTest(unittest.TestCase):

  def test_sync_usages(self):
    client = lf_mcp.McpClient.from_fastmcp(mcp)
    tools = client.list_tools()
    self.assertEqual(len(tools), 1)
    tool_cls = tools['add']
    print(tool_cls.python_definition())
    self.assertEqual(
        tool_cls.python_definition(),
        inspect.cleandoc(
            '''
            Add

            ```python
            class Add:
              """Adds two integers and returns their sum.

              Args:
                a: The first integer.
                b: The second integer.

              Returns:
                The sum of the two integers.
              """
              a: int
              b: int
            ```
            '''
        )
    )
    self.assertEqual(repr(tool_cls), '<tool-class \'Add\'>')
    self.assertEqual(tool_cls.__name__, 'Add')
    self.assertEqual(tool_cls.TOOL_NAME, 'add')
    self.assertEqual(tool_cls(a=1, b=2).input_parameters(), {'a': 1, 'b': 2})
    with client.session() as session:
      self.assertEqual(
          tool_cls(a=1, b=2)(session), 3
      )

  def test_async_usages(self):
    async def _test():
      client = lf_mcp.McpClient.from_fastmcp(mcp)
      tools = client.list_tools()
      self.assertEqual(len(tools), 1)
      tool_cls = tools['add']
      self.assertEqual(tool_cls.__name__, 'Add')
      self.assertEqual(tool_cls.TOOL_NAME, 'add')
      async with client.session() as session:
        self.assertEqual(
            (await tool_cls(a=1, b=2).acall(
                session, returns_call_result=True))
            .structuredContent['result'],
            3
        )
    async_support.invoke_sync(_test)


if __name__ == '__main__':
  unittest.main()




# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base classes for Langfun environment event handlers."""

from langfun.env import interface

Environment = interface.Environment
Sandbox = interface.Sandbox
Feature = interface.Feature


class _SessionEventHandler:
  """Base class for session event handlers."""

  def on_session_start(
      self,
      environment: Environment,
      sandbox: Sandbox,
      session_id: str,
      duration: float,
      error: BaseException | None
  ) -> None:
    """Called when a sandbox session starts.

    Args:
      environment: The environment.
      sandbox: The sandbox.
      session_id: The session ID.
      duration: The time spent on starting the session.
      error: The error that caused the session to start. If None, the session
        started normally.
    """

  def on_session_end(
      self,
      environment: Environment,
      sandbox: Sandbox,
      session_id: str,
      duration: float,
      lifetime: float,
      error: BaseException | None
  ) -> None:
    """Called when a sandbox session ends.

    Args:
      environment: The environment.
      sandbox: The sandbox.
      session_id: The session ID.
      duration: The time spent on ending the session.
      lifetime: The session lifetime in seconds.
      error: The error that caused the session to end. If None, the session
        ended normally.
    """


class _FeatureEventHandler:
  """Base class for feature event handlers."""

  def on_feature_setup(
      self,
      environment: Environment,
      sandbox: Sandbox,
      feature: Feature,
      duration: float,
      error: BaseException | None
  ) -> None:
    """Called when a sandbox feature is setup.

    Args:
      environment: The environment.
      sandbox: The sandbox.
      feature: The feature.
      duration: The feature setup duration in seconds.
      error: The error happened during the feature setup. If None,
        the feature setup performed normally.
    """

  def on_feature_teardown(
      self,
      environment: Environment,
      sandbox: Sandbox,
      feature: Feature,
      duration: float,
      error: BaseException | None
  ) -> None:
    """Called when a sandbox feature is teardown.

    Args:
      environment: The environment.
      sandbox: The sandbox.
      feature: The feature.
      duration: The feature teardown duration in seconds.
      error: The error happened during the feature teardown. If None,
        the feature teardown performed normally.
    """

  def on_feature_teardown_session(
      self,
      environment: Environment,
      sandbox: Sandbox,
      feature: Feature,
      session_id: str,
      duration: float,
      error: BaseException | None
  ) -> None:
    """Called when a feature is teardown with a session.

    Args:
      environment: The environment.
      sandbox: The sandbox.
      feature: The feature.
      session_id: The session ID.
      duration: The feature teardown session duration in seconds.
      error: The error happened during the feature teardown session. If
        None, the feature teardown session performed normally.
    """

  def on_feature_setup_session(
      self,
      environment: Environment,
      sandbox: Sandbox,
      feature: Feature,
      session_id: str | None,
      duration: float,
      error: BaseException | None,
  ) -> None:
    """Called when a feature is setup with a session.

    Args:
      environment: The environment.
      sandbox: The sandbox.
      feature: The feature.
      session_id: The session ID.
      duration: The feature setup session duration in seconds.
      error: The error happened during the feature setup session. If
        None, the feature setup session performed normally.
    """

  def on_feature_housekeep(
      self,
      environment: Environment,
      sandbox: Sandbox,
      feature: Feature,
      counter: int,
      duration: float,
      error: BaseException | None,
      **kwargs,
  ) -> None:
    """Called when a sandbox feature is housekeeping.

    Args:
      environment: The environment.
      sandbox: The sandbox.
      feature: The feature.
      counter: Zero-based counter of the housekeeping round.
      duration: The feature housekeeping duration in seconds.
      error: The error happened during the feature housekeeping. If None, the
        feature housekeeping normally.
      **kwargs: Feature-specific properties computed during housekeeping.
    """


class _SandboxEventHandler(_FeatureEventHandler, _SessionEventHandler):
  """Base class for sandbox event handlers."""

  def on_sandbox_start(
      self,
      environment: Environment,
      sandbox: Sandbox,
      duration: float,
      error: BaseException | None
  ) -> None:
    """Called when a sandbox is started.

    Args:
      environment: The environment.
      sandbox: The sandbox.
      duration: The time spent on starting the sandbox.
      error: The error that caused the sandbox to start. If None, the sandbox
        started normally.
    """

  def on_sandbox_status_change(
      self,
      environment: Environment,
      sandbox: Sandbox,
      old_status: 'Sandbox.Status',
      new_status: 'Sandbox.Status',
      span: float,
  ) -> None:
    """Called when a sandbox status changes.

    Args:
      environment: The environment.
      sandbox: The sandbox.
      old_status: The old sandbox status.
      new_status: The new sandbox status.
      span: Time spent on the old status in seconds.
    """

  def on_sandbox_shutdown(
      self,
      environment: Environment,
      sandbox: Sandbox,
      duration: float,
      lifetime: float,
      error: BaseException | None
  ) -> None:
    """Called when a sandbox is shutdown.

    Args:
      environment: The environment.
      sandbox: The sandbox.
      duration: The time spent on shutting down the sandbox.
      lifetime: The sandbox lifetime in seconds.
      error: The error that caused the sandbox to shutdown. If None, the
        sandbox shutdown normally.
    """

  def on_sandbox_activity(
      self,
      name: str,
      environment: Environment,
      sandbox: Sandbox,
      feature: Feature | None,
      session_id: str | None,
      duration: float,
      error: BaseException | None,
      **kwargs
  ) -> None:
    """Called when a sandbox activity is performed.

    Args:
      name: The name of the sandbox activity.
      environment: The environment.
      sandbox: The sandbox.
      feature: The feature that is associated with the sandbox activity.
      session_id: The session ID.
      duration: The sandbox activity duration in seconds.
      error: The error that caused the sandbox activity to perform. If None,
        the sandbox activity performed normally.
      **kwargs: The keyword arguments of the sandbox activity.
    """

  def on_sandbox_housekeep(
      self,
      environment: Environment,
      sandbox: Sandbox,
      counter: int,
      duration: float,
      error: BaseException | None,
      **kwargs
  ) -> None:
    """Called when a sandbox finishes a round of housekeeping.

    Args:
      environment: The environment.
      sandbox: The sandbox.
      counter: Zero-based counter of the housekeeping round.
      duration: The sandbox housekeeping duration in seconds.
      error: The error that caused the sandbox to housekeeping. If None, the
        sandbox housekeeping normally.
      **kwargs: Sandbox-specific properties computed during housekeeping.
    """


class EventHandler(_SandboxEventHandler):
  """Base class for event handlers of an environment."""

  def on_environment_starting(self, environment: Environment) -> None:
    """Called when the environment is getting started.

    Args:
      environment: The environment.
    """

  def on_environment_start(
      self,
      environment: Environment,
      duration: float,
      error: BaseException | None
  ) -> None:
    """Called when the environment is started.

    Args:
      environment: The environment.
      duration: The environment start duration in seconds.
      error: The error that failed the environment start. If None, the
        environment started normally.
    """

  def on_environment_housekeep(
      self,
      environment: Environment,
      counter: int,
      duration: float,
      error: BaseException | None,
      **kwargs
  ) -> None:
    """Called when the environment finishes a round of housekeeping.

    Args:
      environment: The environment.
      counter: Zero-based counter of the housekeeping round.
      duration: The environment start duration in seconds.
      error: The error that failed the housekeeping. If None, the
        housekeeping succeeded.
      **kwargs: Environment-specific properties computed during housekeeping.
    """

  def on_environment_shutting_down(
      self,
      environment: Environment,
      offline_duration: float,
  ) -> None:
    """Called when the environment is shutting down.

    Args:
      environment: The environment.
      offline_duration: The environment offline duration in seconds.
    """

  def on_environment_shutdown(
      self,
      environment: Environment,
      duration: float,
      lifetime: float,
      error: BaseException | None
  ) -> None:
    """Called when the environment is shutdown.

    Args:
      environment: The environment.
      duration: The environment shutdown duration in seconds.
      lifetime: The environment lifetime in seconds.
      error: The error that caused the environment to shutdown. If None, the
        environment shutdown normally.
    """
